/*
 * Copyright (c) 2011-2012 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE.txt file for licensing information.
 */
package eu.unicore.ucc.ccm;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.CharArrayWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.RandomAccessFile;

import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;

/**
 * Wraps {@link PropertiesConfiguration} with some additional tricks so 
 * the input file is handled in a nicer way for the users. Currently
 * trailing comments are preserved and slashes are not escaped (what 
 * normally {@link PropertiesConfiguration} does).
 * @author K. Benedyczak
 */
public class PropertiesFileHelper
{
	public static final String MAGIC_END = "END.ARTIFICIAL.PROPERTY = END.ARTIFICIAL.PROPERTY";
	
	private String file;
	private PropertiesConfiguration config;
	
	public PropertiesFileHelper(String file)
	{
		this.file = file;
	}

	public PropertiesConfiguration getPropertiesConfiguration() 
			throws IOException, ConfigurationException
	{
		//trick to preserve trailing comments, otherwise stripped by 
		//silly PropertiesConfiguration ;-(
		
		boolean addNL = !checkNLAtEnd();
		BufferedWriter fw = new BufferedWriter(new FileWriter(file, true));
		if (addNL)
			fw.newLine();
		fw.write(MAGIC_END);
		fw.newLine();
		fw.close();
						
		config = new PropertiesConfiguration();
		config.setDelimiterParsingDisabled(true);
		config.setFileName(file);
		config.load();
		return config;
	}

	private boolean checkNLAtEnd() throws IOException
	{
		String lineSep = System.getProperty("line.separator");
		byte []sepBytes = lineSep.getBytes();
		int sepLen = sepBytes.length;
		RandomAccessFile raf = null;
		try
		{
			raf = new RandomAccessFile(file, "r");
			if (raf.length() - sepLen -1 < 0)
				return false;
			raf.seek(raf.length()-sepLen-1);
			byte[] buf = new byte[sepLen];
			raf.read(buf);
			for (int i=0; i<sepLen; i++)
				if (buf[i] != sepBytes[i])
					return false;
		} finally 
		{
			if (raf != null)
				raf.close();			
		}
		return true;
	}
	
	public void close() throws IOException, ConfigurationException
	{
		config.save();
		cleanupSlashes(file);
	}
	
	/**
	 * Unescapes '/' from properties file.
	 * 
	 * @param inputFile
	 */
	private void cleanupSlashes(String inputFile) throws IOException
	{
		BufferedReader br = new BufferedReader(new FileReader(inputFile));
		CharArrayWriter wr = new CharArrayWriter(10240);
		String line;
		String lineSep = System.getProperty("line.separator");
		while ((line = br.readLine()) != null)
		{
			if (line.equals(MAGIC_END))
				continue;
			char[] chars = line.toCharArray();
			int i=0;
			for (; i<chars.length-1; i++)
			{
				if (chars[i] == '\\')
				{
					if (chars[i+1] == '\\')
					{
						i++;
						continue;
					} else if (chars[i+1] == '/')
					{
						wr.append('/');
						i++;
						continue;
					}
				}
				wr.append(chars[i]);
			}
			if (i<chars.length)
				wr.append(chars[i]);
			wr.append(lineSep);
		}
		br.close();
		
		BufferedWriter bw = new BufferedWriter(new FileWriter(inputFile));
		bw.write(wr.toCharArray());
		bw.close();
	}
}
