/*
 * Copyright (c) 2011 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on 16-05-2011
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */
package eu.unicore.ucc.ccm;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.GeneralSecurityException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.Properties;

import org.apache.commons.io.FileUtils;

import de.fzj.unicore.ucc.MessageWriter;

import xmlbeans.eu.unicore.ccm.internal.TrustedCAType;
import xmlbeans.eu.unicore.ccm.internal.ClientProfileDocument.ClientProfile;
import eu.emi.security.authn.x509.helpers.KeyStoreHelper;
import eu.emi.security.authn.x509.impl.CertificateUtils;
import eu.emi.security.authn.x509.impl.KeystoreCredential;
import eu.emi.security.authn.x509.impl.CertificateUtils.Encoding;
import eu.emi.security.authn.x509.impl.X500NameUtils;
import eu.unicore.ccm.ProfileHandler;
import eu.unicore.security.canl.PasswordCallback;
import eu.unicore.security.canl.TruststoreProperties;
import eu.unicore.security.canl.TruststoreProperties.TruststoreType;

/**
 * Updates the trusted CAs
 * @author K. Benedyczak
 */
public class TrustUpdater
{
	public static final String DEFAULT_CA = "demo ca for testing only";
	private char[] password;
	private MessageWriter parent;
	private String workingDir;
	
	public TrustUpdater(MessageWriter parent, String workingDir)
	{
		this.parent = parent;
		this.workingDir = workingDir;
	}
	
	public void processTrustedCAs(ClientProfile profile, boolean update, String propertiesFile) throws CCMException
	{
		List<X509Certificate> certs = fetchTrustedCAs(profile);
		if (certs.size() == 0)
			return;
		
		TruststoreProperties tsProps = initSecurityProperties(propertiesFile);
		TruststoreType type = tsProps.getEnumValue(TruststoreProperties.PROP_TYPE, TruststoreType.class);
		if (type == TruststoreType.keystore)
			processKeystoreTS(profile, certs, update, tsProps);
		else if (type == TruststoreType.directory)
			processDirectoryTS(profile, tsProps, certs);
		else
		{
			parent.message("The currently set truststore type " + type + " is not supported by the " +
				"configuration utility. Therefore the trusted certificates from the profile WERE NOT INSTALLED.");
		}
	}
	
	private void processDirectoryTS(ClientProfile profile, TruststoreProperties tsParams, List<X509Certificate> certs) throws CCMException
	{
		List<String> paths = tsParams.getListOfValues(TruststoreProperties.PROP_DIRECTORY_LOCATIONS);
		String chosenPath = null;
		for (String path: paths)
		{
			File f = new File(path);
			if (f.exists())
				continue;
			if (path.contains(File.separator))
			{
				int split = path.lastIndexOf(File.separator);
				if (split == path.length())
					continue;
				String suffix = path.substring(split+1, path.length());
				path = path.substring(0, split);
				f = new File(path);
				if (f.isDirectory() && f.canWrite())
				{
					if (suffix.equals("*") || suffix.equals("*.*") || suffix.equals("*.pem"))
					{
						chosenPath = path;
						break;
					}
				}
			}
		}
		if (chosenPath == null)
			throw new CCMException("Directory truststore must have at leas one writable directory configured," +
					" with a wildcard expression accepting files with the pem extension " +
					"(e.g. '*.pem') in order to store downloaded certificates.");
		addTrustedCAstoDir(certs, profile, new File(chosenPath));
	}
	
	private void processKeystoreTS(ClientProfile profile, List<X509Certificate> certs, boolean update, 
			TruststoreProperties tsParams) throws CCMException
	{
		String password = tsParams.getValue(TruststoreProperties.PROP_KS_PASSWORD);
		if (password == null)
			password = new String(this.password);
		String truststore = tsParams.getValue(TruststoreProperties.PROP_KS_PATH);
		String type = tsParams.getValue(TruststoreProperties.PROP_KS_TYPE);
		if (type == null)
			try
			{
				type = KeystoreCredential.autodetectType(truststore, password.toCharArray());
			} catch (Exception e1)
			{
				throw new CCMException("Truststore keystore's type is not set and " +
						"its autodetection failed. Try to set it.", e1);
			}

		
		if (update)
			CCMAction.backupFile(truststore, parent);
		else
		{
			try
			{
				FileUtils.copyFileToDirectory(new File(truststore), 
					new File(workingDir));
				truststore = workingDir + File.separator + 
					new File(truststore).getName();
			} catch (IOException e)
			{
				throw new CCMException("Can't copy the " +
						"original truststore", e);
			}
		}
		KeyStore ts = loadTS(truststore, type, password.toCharArray());
		addTrustedCAsToKs(certs, profile, ts);
		saveTS(ts, truststore, password.toCharArray());
	}
	
	
	private KeyStore loadTS(String path, String type, char[] passwd) throws CCMException
	{
		try
		{
			InputStream is = new BufferedInputStream(new FileInputStream(
					path));
			KeyStore ret = KeyStoreHelper.getInstance(type);
			ret.load(is, passwd);
			is.close();
			return ret;
		} catch (GeneralSecurityException e)
		{
			throw new CCMException("Can't load the truststore " + path, e);
		} catch (IOException e)
		{
			throw new CCMException("Can't load the truststore " + path, e);
		}
	}
	
	private void saveTS(KeyStore ts, String path, char[] passwd) throws CCMException
	{
		OutputStream os;
		try
		{
			os = new BufferedOutputStream(new FileOutputStream(path));
			ts.store(os, passwd);
			os.close();
		} catch (GeneralSecurityException e)
		{
			throw new CCMException("Can't save the truststore " + path, e);
		} catch (IOException e)
		{
			throw new CCMException("Can't save the truststore " + path, e);
		}
	}
	
	private List<X509Certificate> fetchTrustedCAs(ClientProfile profile) throws CCMException
	{
		ProfileHandler handler = new ProfileHandler(profile);
		List<X509Certificate> certs;
		try
		{
			certs = handler.downloadCACerts();
		} catch (IOException e)
		{
			throw new CCMException("Can't download trusted CA certificates" +
					" referenced in the profile", e);
		}
		return certs;
	}
	
	private void addTrustedCAsToKs(List<X509Certificate> certs, ClientProfile profile, 
			KeyStore ks) throws CCMException
	{
		if (profile.getRemoveDefaults())
		{
			try
			{
				if (ks.containsAlias(DEFAULT_CA))
					ks.deleteEntry(DEFAULT_CA);
			} catch (KeyStoreException e)
			{
				parent.error("Can't remove the default trusted CA", e);
			}
		}
		TrustedCAType []trustedCAMeta = profile.getTrustedCAArray();
		for (int i=0; i<certs.size(); i++)
		{
			String alias = trustedCAMeta[i].getSuggestedLabel();
			X509Certificate cert = certs.get(i);
			try
			{
				String subject = X500NameUtils.getReadableForm(
						cert.getSubjectX500Principal());
				if (ks.getCertificateAlias(cert) != null)
				{
					parent.verbose("Skipping trusted certificate " + 
							subject + 
							" as it is already present in the truststore.");
					continue;
				}
				if (ks.containsAlias(alias))
				{
					int j=1;
					while(ks.containsAlias(alias + "_" + j))
						j++;
					alias = alias + "_" + j;
				}
				parent.verbose("Will add CA " + subject + " under " + alias + " alias.");
				ks.setCertificateEntry(alias, cert);
				parent.message("Added a new trusted CA: " + cert.getSubjectX500Principal());
			} catch (Exception e)
			{
				throw new CCMException("Can't add trusted CA " +
						"certificate to the keystore", e);
			}
		}
		
	}

	private void addTrustedCAstoDir(List<X509Certificate> certs, ClientProfile profile, 
			File directory) throws CCMException
	{
		TrustedCAType []trustedCAMeta = profile.getTrustedCAArray();
		for (int i=0; i<certs.size(); i++)
		{
			String alias = trustedCAMeta[i].getSuggestedLabel();
			X509Certificate cert = certs.get(i);
			try
			{
				File certFile = new File(directory, alias+".pem");
				int j=1;
				while (certFile.exists())
				{
					certFile = new File(directory, alias+"_"+j+".pem");
				}
				parent.verbose("Will save CA " + X500NameUtils.getReadableForm(
						cert.getSubjectX500Principal()) + 
						" certificate to file " + certFile);
				FileOutputStream fos = new FileOutputStream(certFile);
				CertificateUtils.saveCertificate(fos, cert, Encoding.PEM);
				fos.close();
				parent.message("Added a new trusted CA: " + X500NameUtils.getReadableForm(
						cert.getSubjectX500Principal()));
			} catch (Exception e)
			{
				throw new CCMException("Can't add trusted CA " +
						"certificate to the keystore", e);
			}
		}
		
	}
	

	/**
	 * Make argument relative to properties file, if not absolute
	 * @param file
	 * @return
	 */
	/* TODO - do we still need this somewhere?
	private String makeRelative(String fileName, String propertiesFileName, String opt)
	{
		File file=new File(fileName);
		File propertiesFile = new File(propertiesFileName);
		if (!file.isAbsolute())
		{
			if (propertiesFile!=null && propertiesFile.exists())
				file = new File(propertiesFile.getAbsoluteFile().
						getParent(), fileName);
			return file.getAbsolutePath();
		}
		return fileName;
	}
	*/
	
	private TruststoreProperties initSecurityProperties(String propertiesFile) throws CCMException{
		Properties properties = new Properties();
		try
		{
			properties.load(new FileReader(propertiesFile));
		} catch (Exception e)
		{
			throw new CCMException("Can't load truststore", e);
		}
		TruststoreProperties sp=new TruststoreProperties(properties, null, 
				getPasswordCallback(), TruststoreProperties.DEFAULT_PREFIX);
		return sp;
	}
	

	private PasswordCallback getPasswordCallback() {
		return new PasswordCallback() {
			@Override
			public boolean ignoreProperties() {
				return false;
			}
			
			@Override
			public char[] getPassword(String protectedArtifactType, String protectedArtifactDescription) {
				String r;
				parent.message("Please enter your " + protectedArtifactType + 
						" ("+protectedArtifactDescription + ") password: ");
				try {
					r = new jline.ConsoleReader().readLine(new Character('*'));
				} catch (IOException e) {
					return null;
				}
				r.trim();
				password = r.toCharArray();
				return r.toCharArray();
			}
			
			@Override
			public boolean askForSeparateKeyPassword() {
				return false;
			}
		};
	}
}
