package de.fzj.unicore.ucc.actions;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.ServiceLoader;

import org.apache.commons.cli.OptionBuilder;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.oasisOpen.docs.wsrf.sg2.EntryType;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import xmlbeans.org.oasis.saml2.assertion.AssertionDocument;
import de.fzj.unicore.uas.util.PropertyHelper;
import de.fzj.unicore.ucc.Command;
import de.fzj.unicore.ucc.Constants;
import de.fzj.unicore.ucc.UCC;
import de.fzj.unicore.ucc.UCCOptions;
import de.fzj.unicore.ucc.authn.Authentication;
import de.fzj.unicore.ucc.authn.KeystoreAuthN;
import de.fzj.unicore.ucc.helpers.AttributeAssertionFetcher;
import de.fzj.unicore.ucc.helpers.EndProcessingException;
import de.fzj.unicore.ucc.util.IRegistryFactory;
import de.fzj.unicore.ucc.util.MultiRegistryQuery;
import de.fzj.unicore.ucc.util.VOAttributeFilter;
import de.fzj.unicore.wsrflite.xfire.ClientException;
import de.fzj.unicore.wsrflite.xmlbeans.WSUtilities;
import de.fzj.unicore.wsrflite.xmlbeans.client.IRegistryQuery;
import eu.unicore.samly2.assertion.Assertion;
import eu.unicore.security.etd.TrustDelegation;
import eu.unicore.security.xfireutil.client.SAMLAttributePushOutHandler;
import eu.unicore.util.httpclient.ClientProperties;
import eu.unicore.util.httpclient.IClientConfiguration;

/**
 * Base class for doing something on a UNICORE 6 Grid<br/>
 * Sets up registry and security.
 * 
 * @author schuller
 * @author kstasiak
 */
public abstract class ActionBase extends Command {

	protected IRegistryQuery registry;

	protected IClientConfiguration securityProperties;

	protected String registryURL;

	public static final VOAttributeFilter DEFAULT_ATTRIBUTE_FILTER = VOAttributeFilter.EMPTY_FILTER;
	public static final String PREFERENCE_ARG_VO = "vo";
	public static final String PREFERENCE_ARG_ROLE = "role";
	public static final String PREFERENCE_ARG_UID = "uid";
	public static final String PREFERENCE_ARG_PGID = "pgid";
	public static final String PREFERENCE_ARG_SUP_GIDS = "supgids";
	public static final String PREFERENCE_ARG_ADD_OS_GIDS = "useOSgids";
	public static final String PREFERENCE_ARG = "[see description]";
	public static final String PREFERENCE_ARG_HELP = PREFERENCE_ARG_VO + ":<val>|" + 
			PREFERENCE_ARG_ROLE + ":<val>|" + 
			PREFERENCE_ARG_UID + ":<val>|" +
			PREFERENCE_ARG_PGID + ":<val>|" +
			PREFERENCE_ARG_SUP_GIDS + ":<val1,val2,...>|" +
			PREFERENCE_ARG_ADD_OS_GIDS + ":<true|false>";

	protected String attrAssertionFilename;

	protected String voURL;

	protected List<Element> rawXmlSAMLAssertions;

	protected Assertion rawSAMLAssertion;

	protected VOAttributeFilter attributeFilter;
	
	/**
	 * This variable is set to true if succeeded in setting up SAML PUSH mode.
	 */
	protected boolean useSAMLPush=false;

	/**
	 * creates basic options for setting keystore, etc
	 */
	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();
		createGeneralOptions();
		createSecurityOptions();
		createVOOptions();
	}

	/*
	 * security related (keystore, truststore, ...)
	 */
	@SuppressWarnings("all")
	private void createSecurityOptions(){

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_AUTHN_METHOD_LONG)
				.withDescription("The method used for authentication (default: X509)")
				.withArgName("AuthenticationMethod")
				.hasArg()
				.isRequired(false)
				.create(OPT_AUTHN_METHOD)
				,UCCOptions.GRP_SECURITY);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_DELEGATION_LONG)
				.withDescription("SAML trust delegation assertion")
				.withArgName("TDAssertion")
				.hasArg()
				.isRequired(false)
				.create(OPT_DELEGATION)
				,UCCOptions.GRP_SECURITY);
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_SECURITY_PREFERENCES_LONG)
				.withDescription("User preference regarding choice of VO, role, UNIX login and " +
						"groups which should be used for operation execution. " +
						"The selected values must be allowed for the user. " +
						"The uid option can be also set using " + OPT_USERID + " option. " +
						"Full syntax: "  + PREFERENCE_ARG_HELP)
				.withArgName(PREFERENCE_ARG)
				.hasArgs()
				.isRequired(false)
				.create(OPT_SECURITY_PREFERENCES)
				,UCCOptions.GRP_SECURITY);
	}


	/* 
	 * general (preferences, registries, ...)
	 */
	@SuppressWarnings("all")
	private void createGeneralOptions(){
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_PROPERTIES_LONG)
				.withDescription("Properties file containing your preferences. By default, a file '<userhome>/.ucc/preferences' is checked.")
				.withArgName("Properties")
				.hasArg()
				.isRequired(false)
				.create(OPT_PROPERTIES)
				,UCCOptions.GRP_GENERAL);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_REGISTRY_LONG)
				.withDescription("Comma-separated list of UNICORE registry URLs")
				.withArgName("Registry")
				.hasArg()
				.isRequired(false)
				.create(OPT_REGISTRY)
				,UCCOptions.GRP_GENERAL);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_USERID_LONG)
				.withDescription("The user id to use remotely")
				.withArgName("userID")
				.hasArg()
				.isRequired(false)
				.create(OPT_USERID)
				,UCCOptions.GRP_GENERAL);
	}

	/*
	 * VO related options
	 */
	@SuppressWarnings("all")
	private void createVOOptions(){
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_VO_LONG)
				.withDescription("VO source (use full URL or use \'auto\' to get the URL from the registry)")
				.withArgName("VO_URL")
				.hasArg()
				.isRequired(false)
				.create(OPT_VO)
				,UCCOptions.GRP_VO);
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_ATTRIBUTEASSERTION_LONG)
				.withDescription("Attribute assertion file")
				.withArgName("AssertionFile")
				.hasArg()
				.isRequired(false)
				.create(OPT_ATTRIBUTEASSERTION)
				,UCCOptions.GRP_VO);
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_INCLUDEATTRIBUTES_LONG)
				.withDescription("Regular expressions for included attributes (semicolon-separated); "
						+ "either names or name=value pairs")
						.withArgName("Regexes")
						.hasArg()
						.isRequired(false)
						.create(OPT_INCLUDEATTRIBUTES)
						,UCCOptions.GRP_VO);
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_EXCLUDEATTRIBUTES_LONG)
				.withDescription("Regular expressions for excluded attributes (semicolon-separated); "
						+ "either names or name=value pairs")
						.withArgName("Regexes")
						.hasArg()
						.isRequired(false)
						.create(OPT_EXCLUDEATTRIBUTES)
						,UCCOptions.GRP_VO);
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_VOGROUP_LONG)
				.withDescription("VO or VO and group name (for example: \"/vo/group\")")
				.withArgName("vogroup")
				.hasArg()
				.isRequired(false)
				.create(OPT_VOGROUP)
				,UCCOptions.GRP_VO);
	}

	@Override
	public void process(){
		super.process();
		try{
			initSecurityProperties();
		}catch(Exception ioe){
			error("Problem setting up security.",ioe);
			endProcessing(ERROR_SECURITY);
		}
		if(!skipConnectingToRegistry()){
			initRegistryClient();
		}
		else{
			verbose("Registry connection will be skipped.");
		}
		if(!skipConnectingToVO()){
			connectVO();
		}

		setOutputLocation();
	}

	public IClientConfiguration getClientProperties(){
		return securityProperties;
	}

	/**
	 * connect to the VO server, if required
	 */
	protected void connectVO() {
		voURL = getOption(OPT_VO_LONG, OPT_VO);
		
		if("auto".equals(voURL)){
			List<String> voURLs = new ArrayList<String>();
			try {
				for (EntryType e : registry.listEntries()) {
					EndpointReferenceType mepr = e.getMemberServiceEPR();
					if(mepr.isSetMetadata()
							&& mepr.getMetadata().xmlText().contains("UVOSAssertionQueryService")){
						voURLs.add(mepr.getAddress().getStringValue());
					}
				}
			} catch (Exception ex) {
				error("Error when connection to registry", ex);
				endProcessing(ERROR);
			}
			switch(voURLs.size()){
			case 0:
				error("No VO server URL found in the registry",null);
				endProcessing(ERROR);
				break;
			case 1:
				voURL = voURLs.get(0);
				verbose("Found VO server URL: "+voURL);
				break;
			default:
				error("Multiple VO server URLs found:",null);
				for(String url: voURLs){
					error(" "+url,null);
				}
				endProcessing(ERROR);
			}
		}
		attrAssertionFilename = getOption(OPT_ATTRIBUTEASSERTION_LONG, OPT_ATTRIBUTEASSERTION);

		// we enable SAML push if we have either a UVOS URL ...
		useSAMLPush = voURL!=null && !voURL.equals("");
		// ... or if we have a file containing attribute assertions
		useSAMLPush |= attrAssertionFilename!=null && !attrAssertionFilename.equals("");

		if(!useSAMLPush)return;

		attributeFilter = VOAttributeFilter.build(
				getOption(OPT_INCLUDEATTRIBUTES_LONG, OPT_INCLUDEATTRIBUTES, null),
				getOption(OPT_EXCLUDEATTRIBUTES_LONG, OPT_EXCLUDEATTRIBUTES, null));
		
		if(attrAssertionFilename==null) {
			fetchAssertionFromServer();
		}
		else{
			loadAssertionFromFile();
		}

		addSAMLOutHandler((ClientProperties)securityProperties);
	}

	/**
	 * Function that returns true if the connection to registry is not required.
	 * By default, this checks a property "contact-registry". If this is not
	 * set, returns <code>true</code>, otherwise the value is parsed as a boolean. 
	 * Override if needed.
	 */
	protected boolean skipConnectingToRegistry(){
		return !Boolean.parseBoolean(properties.getProperty("contact-registry", "true"));
	}

	/**
	 * return <code>true</code> if the command REQUIRES an option or configuration value for the
	 * registry 
	 */
	protected boolean requireRegistry(){
		return true;
	}
	
	/**
	 * Function that returns true if VO connection is not required.
	 * By default always returns false; override if needed.
	 */
	protected boolean skipConnectingToVO(){
		return false;
	}

	/**
	 * set the security properties to be used
	 * @param sec - the security properties
	 */
	public void setSecurityProperties(IClientConfiguration sec){
		this.securityProperties=sec;
	}

	/**
	 * initialise the properties for outgoing client calls (if not already set)
	 * 
	 * @throws IOException
	 */
	protected void initSecurityProperties()throws Exception{
		if(securityProperties!=null)return;
		loadUserProperties();
		//load AuthN method
		String authNMethod=getOption(OPT_AUTHN_METHOD_LONG, OPT_AUTHN_METHOD, KeystoreAuthN.X509);
		Authentication authN=UCC.getAuthNMethod(authNMethod);
		if(authN==null){
			throw new Exception("No such authentication method: "+authNMethod);
		}
		securityProperties=authN.initSecurityProperties(properties);
		securityProperties.getETDSettings().setExtendTrustDelegation(true);
		handleExtraConfigForOutHandlers();
	}

	
	//TODO - some handlers need per-handler properties! Should remove this code and 
	//add a generic code to ClientProperties.
	private void handleExtraConfigForOutHandlers(){
		String[]configHandlers=securityProperties.getOutHandlerClassNames();
		for(String h: configHandlers){
			//check if we have additional config options
			String propKey=h+".";
			PropertyHelper ph=new PropertyHelper(properties,propKey);
			for(Map.Entry<String, String>entry: ph.getFilteredMap().entrySet()){
				String key=entry.getKey().substring(propKey.length());
				String value=entry.getValue();
				//value might be sensitive so do not log it
				verbose("Setting handler property <"+key+">");
				securityProperties.getExtraSettings().setProperty(key,value);
			}
		}
	}

	
	/**
	 * add any extra attributes to the security properties. Called from
	 * the {@link #initSecurityProperties()}.<br/>
	 * The base implementation adds the requested user ID, if given by
	 * the {@link Constants#OPT_USERID} option
	 * @param sp - the security properties currently being created
	 */
	protected void setupExtraAttributes(ClientProperties sp){
		String[] vals = getCommandLine().getOptionValues(OPT_SECURITY_PREFERENCES);
		if (vals == null) {
			String fromFile = properties.getProperty(OPT_SECURITY_PREFERENCES_LONG);
			if (fromFile != null)
				vals = fromFile.split("( )+");
		}
		if (vals != null) {
			for (String val: vals)
				parseSecurityPreference(val, sp);
		}
		
		//xlogin have also its own option, which overrides generic syntax setting
		String userID=getOption(OPT_USERID_LONG, OPT_USERID);
		if(userID!=null){
			sp.getETDSettings().getRequestedUserAttributes2().put("xlogin", new String[]{userID});
		}
	}
	
	protected void parseSecurityPreference(String pref, ClientProperties sp) {
		if (pref.startsWith(PREFERENCE_ARG_UID + ":")) {
			String val = pref.substring(PREFERENCE_ARG_UID.length() + 1);
			sp.getETDSettings().getRequestedUserAttributes2().put(
				"xlogin", new String[]{val});
		} else if (pref.startsWith(PREFERENCE_ARG_PGID + ":")) {
			String val = pref.substring(PREFERENCE_ARG_PGID.length() + 1);
			sp.getETDSettings().getRequestedUserAttributes2().put(
				"group", new String[]{val});
		} else if (pref.startsWith(PREFERENCE_ARG_SUP_GIDS + ":")) {
			String val = pref.substring(PREFERENCE_ARG_SUP_GIDS.length() + 1);
			String vals[] = val.split(",");
			sp.getETDSettings().getRequestedUserAttributes2().put(
				"supplementaryGroups", vals);
		} else if (pref.startsWith(PREFERENCE_ARG_ADD_OS_GIDS + ":")) {
			String val = pref.substring(PREFERENCE_ARG_ADD_OS_GIDS.length() + 1);
			if (!val.equals("true")&&!val.equals("false")) {
				error("Value of the " + PREFERENCE_ARG_ADD_OS_GIDS + 
					" preferrence must be 'true' or 'false', but not '" + 
					val + "'", null);
				endProcessing(ERROR_CLIENT);
			}
			sp.getETDSettings().getRequestedUserAttributes2().put(
				"addDefaultGroups", new String[]{val});
		} else if (pref.startsWith(PREFERENCE_ARG_VO + ":")) {
			String val = pref.substring(PREFERENCE_ARG_VO.length() + 1);
			sp.getETDSettings().getRequestedUserAttributes2().put(
				"selectedVirtualOrganisation", new String[]{val});
		} else if (pref.startsWith(PREFERENCE_ARG_ROLE + ":")) {
			String val = pref.substring(PREFERENCE_ARG_ROLE.length() + 1);
			sp.getETDSettings().getRequestedUserAttributes2().put(
				"role", new String[]{val});
		} else {
			error("Wrong value '" + pref + "' of the option -" + OPT_SECURITY_PREFERENCES + 
				", must have the following format: " + PREFERENCE_ARG_HELP, null);
			endProcessing(ERROR_CLIENT);
		}
	}

	/**
	 * reads trust delegation tokens, if given by
	 * the "-D" option
	 * 
	 * @return List<TrustDelegation>
	 */
	protected List<TrustDelegation>getTDTokens()throws IOException{
		String tdPath=getOption(OPT_DELEGATION_LONG, OPT_DELEGATION);
		if(tdPath==null)return null;
		File tdFile=new File(tdPath);
		try{
			AssertionDocument td=AssertionDocument.Factory.parse(tdFile);
			verbose("Read TD tokens from "+tdPath);
			List<TrustDelegation>res=new ArrayList<TrustDelegation>();
			res.add(new TrustDelegation(td));
			return res;
		}catch(Exception xe){
			IOException e= new IOException("Attempted to create trust delegation token from invalid XML.");
			e.initCause(xe);
			throw e;
		}
	}

	/**
	 * setup the registry client to be used. The registry URL is taken from the 
	 * commandline or the properties file. If no registry is configured, registry connction
	 * will not be done. This may or not lead to issues later, depending on the type of command
	 */
	protected void initRegistryClient(){
		registryURL=getCommandLine().getOptionValue(OPT_REGISTRY, properties.getProperty(OPT_REGISTRY_LONG));
		if(registryURL==null || registryURL.trim().length()==0){
			verbose("No registry is configured.");
			if(requireRegistry()){
				throw new EndProcessingException(ERROR, "A registry is required: please use the '-r' option " +
						"or configuration entry to define the registry to be used.");
			}
			return;
		}
		verbose("Registry = "+registryURL);
		//accept list of registries either comma- or space-separateed
		String[] urls=registryURL.split("[, ]");
		if(urls.length>1){
			MultiRegistryQuery erc= new MultiRegistryQuery(this);
			for(String url: urls){
				if(url.trim().length()==0)continue;
				verbose("Registry = "+url);
				try{
					IRegistryQuery c=makeRegistry(url);
					erc.addRegistry(c);
				}catch(Exception e){
					error("Cannot contact registry <"+url+">",null);
				}	
			}
			registry=erc;
		}
		else{
			try{
				registry=makeRegistry(registryURL);
				testRegistryConnection();
			}catch(Exception e){
				error("Cannot contact registry",e);
				endProcessing();
			}
		}
	}

	protected IRegistryQuery makeRegistry(String url)throws Exception{
		ServiceLoader<IRegistryFactory>sl=ServiceLoader.load(IRegistryFactory.class);
		for(IRegistryFactory f: sl){
			try{
				IRegistryQuery q=f.create(url, securityProperties, this);
				if(q!=null)return q;
			}catch(Throwable ex){
				error("Problem trying to create registry client using "+f.getClass().getName(),ex);
			}
		}
		return null;
	}

	/**
	 * checks if the registry can be contacted
	 * 
	 */
	protected void testRegistryConnection(){
		try{
			verbose("Checking registry connection.");
			String status=registry.getConnectionStatus();
			verbose("Registry connection status: "+status);
		}catch(ClientException e){
			error("Cannot contact registry",e);
			endProcessing(ERROR_CLIENT);
		}
	}

	/**
	 * in case an EPR is not taken from a registry, we might not know the Server DN,
	 * and thus cannot do trust delegation. In this case we try to find
	 * a service on the same server (e.g. a TSF) that has published its DN
	 *  
	 * @param epr - an {@link EndpointReferenceType}
	 * @return the DN for that EPR
	 */
	protected String findServerName(EndpointReferenceType epr)throws Exception{
		//first, check if server name is already in the EPR...
		String dn=WSUtilities.extractServerIDFromEPR(epr);
		if(dn!=null){
			return dn;
		}
		//otherwise find a matching service in the registry
		String url=epr.getAddress().getStringValue();
		if(url.contains("/services/"))url=url.substring(0,url.indexOf("/services"));
		verbose("Checking for services at "+url);
		for(EntryType entry:registry.listEntries()){
			if(entry.getMemberServiceEPR().getAddress().getStringValue().startsWith(url)){
				dn=WSUtilities.extractServerIDFromEPR(entry.getMemberServiceEPR());
				if(dn!=null){
					return dn;
				}
			}
		}
		return null;
	}

	/**
	 * loads an assertion from attrAssertionFilename.
	 */
	private void loadAssertionFromFile(){
		File assertionFile = new File(attrAssertionFilename);
		if(!assertionFile.exists() || !assertionFile.canRead()){
			error("Assertion file "+assertionFile.getAbsolutePath()+" does not exist or is not readable.",null);
			endProcessing(ERROR_CLIENT);
		}
		try{
			AssertionDocument a = AssertionDocument.Factory.parse(assertionFile);
			assertionInstallCommon(new Assertion(a));
		}catch(Exception ex){
			error("Error loading assertion from "+assertionFile.getAbsolutePath(),ex);
			endProcessing(ERROR_CLIENT);
		}
	}

	/**
	 * Fetches an assertion from server and stores it 
	 * in {@link #rawSAMLAssertion this.rawSAMLAssertion}
	 * and {@link #rawXmlSAMLAssertions this.rawXmlSAMLAssertions}<br/>
	 * 
	 * Quits in case of errors
	 */
	protected void fetchAssertionFromServer() {
		String group=getOption(OPT_VOGROUP_LONG, OPT_VOGROUP, null);
			
		try {
			AttributeAssertionFetcher fetcher = new AttributeAssertionFetcher(securityProperties, voURL);
			Assertion a = fetcher.fetchAssertion(group, attributeFilter);
			assertionInstallCommon(a);
		} catch (Exception ex) {
			error(ex.getMessage(), ex);
			endProcessing(ERROR_SECURITY);
		}
	}
	
	/**
	 * installs assertion: sets local variables, and sets up security context so 
	 * out handler can pick it up.
	 * @param assertion
	 * @throws JDOMException
	 * @throws IOException
	 */
	private void assertionInstallCommon(Assertion assertion) throws JDOMException, IOException
	{
		rawSAMLAssertion = assertion;
		rawXmlSAMLAssertions = Arrays.asList(convertAssertion(rawSAMLAssertion));
		if(rawXmlSAMLAssertions!=null){
			Map<String, Object> secContext = securityProperties.getExtraSecurityTokens();
			secContext.put(SAMLAttributePushOutHandler.PUSHED_RAW_ASSERTIONS, rawXmlSAMLAssertions);
		}
	}

	/**
	 * Adds SAML handler which will add SAML assertion to the outgoing messages.
	 * @param sp
	 */
	private void addSAMLOutHandler(ClientProperties sp){
		if (!useSAMLPush)
			return;
		String[] outHandlers = sp.getOutHandlerClassNames();
		for (String h: outHandlers)
			if (h.equals(SAMLAttributePushOutHandler.class.getName()))
				return;
		outHandlers = Arrays.copyOf(outHandlers, outHandlers.length + 1);
		outHandlers[outHandlers.length-1] = SAMLAttributePushOutHandler.class.getName();
		sp.setOutHandlerClassNames(outHandlers);
	}
	/**
	 * Converts assertion to XML Element
	 * @param a assertion
	 * @return XML Element
	 * @throws JDOMException
	 * @throws IOException
	 */
	private static Element convertAssertion(Assertion a) throws JDOMException, IOException{
		return new SAXBuilder().build(a.getXML().newInputStream()).detachRootElement();
	}
}