package de.fzj.unicore.ucc.actions;

import java.io.File;
import java.io.FileWriter;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import javax.security.auth.x500.X500Principal;

import org.apache.commons.cli.OptionBuilder;
import org.unigrids.services.atomic.types.ProtocolType;
import org.unigrids.x2006.x04.services.fts.SummaryType;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.TransferControllerClient;
import de.fzj.unicore.ucc.helpers.Location;
import de.fzj.unicore.ucc.util.ProgressBar;
import de.fzj.unicore.ucc.util.UnitParser;
import de.fzj.unicore.wsrflite.xmlbeans.BaseFault;
import de.fzj.unicore.wsrflite.xmlbeans.WSUtilities;

/**
 * Copy a file from a remote sms to another remote SMS <br/>
 *
 * (if source and target are the same SMS, the 
 * simple StorageManagement copy operation is used)
 */
public class CopyFile extends FileOperation {

	protected Location sourceDesc;
	protected Location targetDesc;
	protected TransferControllerClient tcc;
	
	protected StorageClient sms;
	
	protected boolean synchronous;
	
	protected long remoteSize=-1;

	protected boolean forceRemoteCopy=false;
	
	protected boolean reliable=false;
	
	protected String scheduled=null;
	
	public static final String OPT_RELIABLE_LONG="reliable";
	public static final String OPT_RELIABLE="R";
	
	private static String lastFTAddress;
	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_MODE_LONG)
				.withDescription("Asynchronous mode, writes the transfer ID to a file.")
				.isRequired(false)
				.create(OPT_MODE)
				);
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_FORCE_REMOTE_MODE_LONG)
				.withDescription("Always copy remotely even if files are on the same storage")
				.isRequired(false)
				.create(OPT_FORCE_REMOTE_MODE)
				);
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_SCHEDULED_LONG)
				.withDescription("Schedule the transfer for a specific time (in ISO8601 format)")
				.isRequired(false)
				.hasArg()
				.create(OPT_SCHEDULED)
				);
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_RELIABLE_LONG)
				.withDescription("Enable server-side reliable transfer mode")
				.isRequired(false)
				.create(OPT_RELIABLE)
				);
	}

	@Override
	public void process(){
		super.process();
		lastFTAddress=null;
		
		synchronous=!getBooleanOption(OPT_MODE_LONG, OPT_MODE);
		verbose("Synchronous transfer = "+synchronous);
		
		forceRemoteCopy=getBooleanOption(OPT_FORCE_REMOTE_MODE_LONG, OPT_FORCE_REMOTE_MODE);
		verbose("Always copying remotely = "+forceRemoteCopy);
		
		scheduled=getOption(OPT_SCHEDULED_LONG, OPT_SCHEDULED);
		if(scheduled!=null){
			scheduled=UnitParser.convertDateToISO8601(scheduled);
			verbose("Will schedule transfer for "+scheduled);
		}
		reliable=getBooleanOption(OPT_RELIABLE_LONG, OPT_RELIABLE);
		if(reliable)verbose("Will request reliable mode");
		
		setOutputLocation();
		sourceDesc=new Location(source,registry,securityProperties,this);
		targetDesc=new Location(target,registry,securityProperties,this);
		
		EndpointReferenceType targetEPR=EndpointReferenceType.Factory.newInstance();
		targetEPR.addNewAddress().setStringValue(targetDesc.getSmsEpr());
		try{
			String dn=findServerName(targetEPR);
			if(dn!=null){
				securityProperties.getETDSettings().setReceiver(new X500Principal(dn));
				verbose("Delegating to "+dn);
			}
			
			sms=new StorageClient(targetDesc.getSmsEpr(),targetEPR,securityProperties);
			verbose("SMS "+targetDesc.getSmsEpr()
					+" contacted, server reply: servertime="+sms.getCurrentTime());
			if(timing){
				startTime=System.currentTimeMillis();
			}
		}catch(Exception ex){
			error("Can't setup copy-file.",ex);
			endProcessing(1);
		}
		if(!forceRemoteCopy && sourceDesc.getSmsEpr().equalsIgnoreCase(targetDesc.getSmsEpr())){
			smsCopyFile();
		}
		else copyFile();
		if(timing){
			endTime=System.currentTimeMillis();
			long duration=endTime-startTime;
			double rate=(double)remoteSize/(double)duration;
			message("Rate: "+numberFormat.format(rate)+ " kB/sec.");
		}
	}
	
	/**
	 * perform the remote copy, and cleanup the file transfer resource (if not in async mode)
	 */
	protected void copyFile(){
		try{
			checkProtocols();
			verbose("Initiating FetchFile on SMS=<"+sms.getEPR().getAddress().getStringValue()+">," +
					"receiving file <"+source+">, writing to "+targetDesc.getName());
			tcc=sms.fetchFile(sourceDesc.getUnicore6URI(), targetDesc.getName(),getExtraParams(),reliable);
			String ftAddress=tcc.getEPR().getAddress().getStringValue();
			lastFTAddress=ftAddress;
			verbose("Have filetransfer instance: "+ftAddress);
			if(!synchronous){
				writeIDFile();
				return;
			}
			waitForCompletion();
		}catch(BaseFault f){
			error("Can't copy file.",f);
			endProcessing(ERROR_SERVER);
		}catch(Exception e){
			error("Can't copy file.",e);
			endProcessing(ERROR_CLIENT);
		}
		finally{
			if(synchronous){
				try{ tcc.destroy(); }
				catch(Exception e1){
					error("Could not destroy the filetransfer client",e1);
				}
			}
		}
	}
	
	protected Map<String,String>getExtraParams(){
		Map<String, String>params=new HashMap<String, String>();
		if(scheduled!=null){
			params.put("scheduledStartTime",scheduled);
		}
		return params;
	}
	
	/**
	 * wait until complete or failed
	 * @throws Exception
	 */
	protected void waitForCompletion()throws Exception{
		long transferred=-1;
		remoteSize=tcc.getSize();
		ProgressBar p=new ProgressBar(sourceDesc.getName(),remoteSize,this);
		tcc.setUpdateInterval(-1);
		SummaryType.Enum status=SummaryType.UNDEFINED;
		do{
			Thread.sleep(1000);
			remoteSize=tcc.getSize();
			transferred=tcc.getTransferredBytes();
			p.setSize(remoteSize);
			p.updateTotal(transferred);
			status=tcc.getStatusSummary();
		}while(!SummaryType.FAILED.equals(status) && !tcc.isComplete() );
		remoteSize=tcc.getSize();
		transferred=tcc.getTransferredBytes();
		p.finish();
		if(SummaryType.FAILED.equals(status)){
			String desc=tcc.getStatus();
			throw new Exception("File transfer "+desc);
		}
	}
	
	protected void smsCopyFile(){
		try{
			verbose("Copy on remote storage: "+sourceDesc.getName()+"->"+targetDesc.getName());
			sms.copy(sourceDesc.getName(), targetDesc.getName());
		}catch(BaseFault f){
			error("Can't copy file.",f);
			endProcessing(ERROR_SERVER);
		}
		catch(Exception e){
			error("Can't copy file.",e);
			endProcessing(ERROR_CLIENT);
		}
	}
	
	@Override
	public String getName() {
		return "copy-file";
	}

	@Override
	public String getSynopsis() {
		return "Copies a file from a remote location '-s' and " +
				"writes it to the remote storage specified by the " +
				"'-t' option.";
	}
	
	@Override
	public String getDescription(){
		return "copy remote files";
	}
	
	@Override
	public String getArgumentList(){
		return "";
	}
	

	protected void writeIDFile(){
		try{
			File dump=new File(output,WSUtilities.extractResourceID(tcc.getEPR())+".transfer");
			FileWriter fw=new FileWriter(dump);
			fw.append(tcc.getEPR().toString()+"\n");
			fw.close();
			verbose("Wrote transfer descriptor to "+dump.getAbsolutePath());
			message(dump.getAbsolutePath());
		}
		catch(Exception e){
			error("Could not write job ID file.",e);
		}
	}
	
	protected void checkProtocols()throws Exception{
		if(preferredProtocols.size()>1){
			for(ProtocolType.Enum p: preferredProtocols){
				//only check source SMS, assuming it will do the checking of the target SMS 
				if(!Arrays.asList(sms.getSupportedProtocols()).contains(p))continue;
				sourceDesc.setProtocol(p.toString());
				verbose("Using preferred protocol: "+p);
				break;
			}
		}
	}

	@Override
	public String getCommandGroup(){
		return "Data management";
	}
	
	static String getLastFTAddress(){
		return lastFTAddress;
	}
}
