package de.fzj.unicore.ucc.actions;

import java.util.Calendar;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.commons.cli.OptionBuilder;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.OperatingSystemType;
import org.unigrids.services.atomic.types.TextInfoType;
import org.unigrids.x2006.x04.services.tsf.CreateTSRDocument;
import org.unigrids.x2006.x04.services.tsf.TargetSystemDescriptionType;
import org.unigrids.x2006.x04.services.tsf.TargetSystemFactoryPropertiesDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.TargetSystemFactory;
import de.fzj.unicore.uas.client.TSFClient;
import de.fzj.unicore.uas.client.TSSClient;
import de.fzj.unicore.ucc.IServiceInfoProvider;
import de.fzj.unicore.ucc.helpers.TargetSystemFactoryLister;
import de.fzj.unicore.ucc.helpers.TargetSystemFactoryLister.TSFFilter;
import eu.unicore.util.httpclient.IClientConfiguration;

/**
 * creates an TSS instance
 * 
 * @since 1.6.0
 * @author schuller
 */
public class CreateTSS extends ActionBase implements IServiceInfoProvider{

	/**
	 * the initial lifetime (in days) for newly created TSSs
	 */
	private int initialLifeTime;

	/**
	 * factory URL to use
	 */
	private String factoryURL;

	/**
	 * site where to create the TSS
	 */
	private String siteName;

	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_LIFETIME_LONG)
				.withDescription("Initial lifetime (in days) for created target systems.")
				.withArgName("Lifetime")
				.hasArg()
				.isRequired(false)
				.create(OPT_LIFETIME)
				);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_SITENAME_LONG)
				.withDescription("Name of the site")
				.withArgName("Site")
				.hasArg()
				.isRequired(false)
				.create(OPT_SITENAME)
				);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_FACTORY_LONG)
				.withDescription("Factory URL")
				.withArgName("Factory")
				.hasArg()
				.isRequired(false)
				.create(OPT_FACTORY)
				);
	}

	@Override
	public String getName(){
		return "create-tss";
	}

	@Override
	public String getArgumentList(){
		return "[param1=value1,...]";
	}

	@Override
	public String getSynopsis(){
		return "Creates a target system service instance. " +
				"The TargetSystemFactory to be used can be specified. Optionally, " +
				"additional parameters can be given.";

	}
	@Override
	public String getDescription(){
		return "create a target system service instance";
	}

	@Override
	public void process() {
		super.process();

		initialLifeTime=getNumericOption(OPT_LIFETIME_LONG, OPT_LIFETIME, -1);
		if(initialLifeTime>0){
			verbose("New TSSs will have a lifetime of <"+initialLifeTime+"> days.");
		}else{
			verbose("Using site default for TSS lifetime.");
		}

		factoryURL=getOption(OPT_FACTORY_LONG, OPT_FACTORY);
		TSFClient tsf=null;

		if(factoryURL==null){
			siteName=getOption(OPT_SITENAME_LONG, OPT_SITENAME);
			if(siteName!=null){
				verbose("Looking for factory at site <"+siteName+">");
			}
			else{
				verbose("No factory specified, will choose one from registry.");
			}
		}
		else{
			verbose("Requiring factory at <"+factoryURL+">");
		}

		TargetSystemFactoryLister tsfl=new TargetSystemFactoryLister(registry,securityProperties,new Filter(),this);
		List<TSFClient>list=tsfl.list();
		if(list.size()>0){
			tsf=list.get(0);
			factoryURL=tsf.getUrl();
			verbose("Using factory at <"+factoryURL+">");
		}
		
		if(tsf==null){
			error("No suitable target system factory available!",null);
			endProcessing(ERROR);
		}

		try{
			TSSClient tss=tsf.createTSS(getCreateTSRDocument());
			String addr=tss.getEPR().getAddress().getStringValue();
			message(addr);
			setLastTSSAddress(addr);
		}catch(Exception ex){
			error("Could not create storage",ex);
			endProcessing(ERROR);
		}
	}
	
	protected CreateTSRDocument getCreateTSRDocument(){
		CreateTSRDocument in=CreateTSRDocument.Factory.newInstance();
		in.addNewCreateTSR();
		if(initialLifeTime>0){
			in.getCreateTSR().addNewTerminationTime().setCalendarValue(getTermTime());
		}
		
		if(getCommandLine().getArgs().length>1){
			TargetSystemDescriptionType desc=in.getCreateTSR().addNewTargetSystemDescription();
			//other parameters from the cmdline as key=value
			for(int i=1;i<getCommandLine().getArgs().length;i++){
				String arg=getCommandLine().getArgs()[i];
				String[]sp=arg.split("=",2);
				TextInfoType prop=desc.addNewTextInfo();
				prop.setName(sp[0]);
				prop.setValue(sp[1]);
				verbose("Have parameter : "+arg);
			}
			lastCreateTSRDoc=in;
		}
		return in;
	}

	protected Calendar getTermTime(){
		Calendar c = Calendar.getInstance();
		c.add(Calendar.DATE, initialLifeTime);
		return c;
	}

	@Override
	public String getCommandGroup(){
		return "General";
	}
	
	public QName getQName(){
		return TargetSystemFactory.TSF_PORT;
	}

	public String getServiceName(){
		return "TargetSystemFactory";
	}

	public String getServiceDetails(EndpointReferenceType epr, IClientConfiguration securityProperties){
		try{
			TSFClient tsf=new TSFClient(epr.getAddress().getStringValue(),epr,securityProperties);
			StringBuilder sb=new StringBuilder();
			String name=tsf.getResourcePropertiesDocument().getTargetSystemFactoryProperties().getName();
			sb.append(name).append(": ");
			try{
				int cpus=getTotalCPUs(tsf.getResourcePropertiesDocument());
				if(cpus>0)sb.append(cpus).append(" CPUs");
			}catch(Exception ex){}
			try{
				OperatingSystemType[]os=tsf.getOperatingSystems();
				if(os.length>0)sb.append(" ( ");
				for(OperatingSystemType o: os){
					sb.append(o.getOperatingSystemType().getOperatingSystemName().toString()).append(" ");
				}
				if(os.length>0)sb.append(")");
			}catch(Exception ex){}
			return sb.toString();
		}catch(Exception ex){}
		return null;
	}
	
	private int getTotalCPUs(TargetSystemFactoryPropertiesDocument rpDoc){
		int cpus=-1;
		if(rpDoc.getTargetSystemFactoryProperties().getTotalCPUCount()!=null){
			cpus=(int)rpDoc.getTargetSystemFactoryProperties().getTotalCPUCount().getRangeArray()[0].getUpperBound().getDoubleValue();
		}
		else if(rpDoc.getTargetSystemFactoryProperties().getIndividualCPUCount()!=null){
			cpus=(int)rpDoc.getTargetSystemFactoryProperties().getIndividualCPUCount().getRangeArray()[0].getUpperBound().getDoubleValue();
			if(rpDoc.getTargetSystemFactoryProperties().getTotalResourceCount()!=null){
				int nodes=(int)rpDoc.getTargetSystemFactoryProperties().getTotalResourceCount().getRangeArray()[0].getUpperBound().getDoubleValue();
				cpus*=nodes;
			}
		} 
		return  cpus;
	}
	

	private static String lastTargetSystemAddress;

	public static String getLastTargetSystemAddress() {
		return lastTargetSystemAddress;
	}
	
	private static CreateTSRDocument lastCreateTSRDoc;

	static CreateTSRDocument getLastCreateTSRDoc() {
		return lastCreateTSRDoc;
	}

	protected static void setLastTSSAddress(String lastAddress) {
		CreateTSS.lastTargetSystemAddress = lastAddress;
	}

	
	private class Filter implements TSFFilter{

		@Override
		public boolean accept(EndpointReferenceType epr) {
			if(factoryURL==null)return true;
			else{
				return epr.getAddress().getStringValue().equalsIgnoreCase(factoryURL);
			}
		}

		@Override
		public boolean accept(TSFClient smf) {
			return checkName(smf); 
		}

		private boolean checkName(TSFClient smf) {
			if(siteName==null)return true;
			String url=smf.getEPR().getAddress().getStringValue();
			return url.contains("/"+siteName+"/");
		}
	}

}
