package de.fzj.unicore.ucc.actions;

import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.apache.commons.cli.OptionBuilder;
import org.unigrids.services.atomic.types.ProtocolType;

import de.fzj.unicore.ucc.StorageConstants;

/**
 * A basic fileoperation that has a source and target option
 * 
 * @author schuller
 */
public abstract class FileOperation extends ActionBase implements StorageConstants{

	protected String source;
	protected String target;
	
	//index of first byte to process
	protected Long startByte;
	
	//index of last byte to process
	protected Long endByte;

	/**
	 * List of preferred protocols, which can be set from the commandline, or from the preferences
	 * file. At least the BFT protocol is guaranteed to be in this list.
	 */
	protected final List<ProtocolType.Enum> preferredProtocols=new ArrayList<ProtocolType.Enum>();

	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();
		if(needSourceTargetOpts()){
			getOptions().addOption(OptionBuilder.withLongOpt(OPT_SOURCE_LONG)
					.withDescription("Data source")
					.withArgName("Source")
					.hasArg()
					.isRequired(true)
					.create(OPT_SOURCE)
					);
			getOptions().addOption(OptionBuilder.withLongOpt(OPT_TARGET_LONG)
					.withDescription("Data target")
					.withArgName("Target")
					.hasArg()
					.isRequired(true)
					.create(OPT_TARGET)
					);
		}
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_PROTOCOLS_LONG)
				.withDescription("Preferred Protocols")
				.withArgName("ProtocolList")
				.hasArgs()
				.isRequired(false)
				.create(OPT_PROTOCOLS)
		);
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_BYTES_LONG)
				.withDescription("Byte range")
				.withArgName("ByteRange")
				.hasArg()
				.isRequired(false)
				.create(OPT_BYTES)
		);
	}

	protected boolean needSourceTargetOpts(){
		return true;
	}
	
	@Override
	public void process() {
		super.process();
		initPreferredProtocols();
		source=getCommandLine().getOptionValue(OPT_SOURCE);
		target=getCommandLine().getOptionValue(OPT_TARGET);
		initRange();
	}

	/**
	 * get the list of protocols. The first matching one (supported by the 
	 * SMS) will be used. This can be set using a commandline option or 
	 * preferences entry
	 * 
	 * @return ProtocolType.Enum[]
	 */
	private void initPreferredProtocols(){
		String[] protocols=getCommandLine().getOptionValues(OPT_PROTOCOLS);
		preferredProtocols.clear();
		preferredProtocols.addAll(getPreferredProtocols(protocols, properties));
		verbose("Using preferred protocols: "+preferredProtocols);
	}
	
	/**
	 * create a list of preferred protocols from the given space separated string.
	 * If this is empty, the properties are checked for a "protocols" property
	 * 
	 * @param protocols
	 * @param properties
	 * @return list of preferred protocols
	 */
	public static List<ProtocolType.Enum>getPreferredProtocols(String[] protocols, Properties properties)throws IllegalArgumentException{
		List<ProtocolType.Enum> result=new ArrayList<ProtocolType.Enum>();
		if(protocols==null || protocols.length==0){
			if(properties!=null){
				//get from preferences
				String protoP=properties.getProperty(OPT_PROTOCOLS_LONG,"BFT");
				protocols=protoP.split(" +");
			}
			if(protocols==null || protocols.length==0){
				protocols=new String[]{"BFT"};
			}
		}
		for(String p: protocols){
			if(p!=null){
				ProtocolType.Enum x=ProtocolType.Enum.forString(p);
				if(x!=null){
					result.add(x);
				}
				else throw new IllegalArgumentException("'"+p+"' does not denote a valid protocol.");
			}
		}
		if(!result.contains(ProtocolType.BFT)){
			result.add(ProtocolType.BFT);
		}
		return result;
	}

	
	protected void initRange(){
		String bytes=getOption(OPT_BYTES_LONG, OPT_BYTES);
		if(bytes==null)return;
		String[]tokens=bytes.split("-");
		try{
			String start=tokens[0];
			String end=tokens[1];
			if(start.length()>0){
				startByte=Long.parseLong(start);
				endByte=Long.MAX_VALUE;
			}
			if(end.length()>0){
				endByte=Long.parseLong(end);
				if(startByte==null){
					startByte=Long.valueOf(0l);
				}
				
			}
		}catch(Exception e){
			throw new IllegalArgumentException("Could not parse byte range "+bytes);
		}
	}
}
