package de.fzj.unicore.ucc.actions;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Date;

import javax.security.auth.x500.X500Principal;

import org.apache.commons.cli.OptionBuilder;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.client.TSSClient;
import de.fzj.unicore.ucc.helpers.TargetSystemLister;
import de.fzj.unicore.wsrflite.xmlbeans.WSUtilities;
import eu.unicore.security.UnicoreSecurityFactory;
import eu.unicore.security.etd.DelegationRestrictions;
import eu.unicore.security.etd.TrustDelegation;

/**
 * issue a trust delegation assertion
 * 
 * @author schuller
 */
public class IssueTD extends ActionBase {
	
	private String subject;
	private int validity;
	private String target;
	
	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();
		getOptions().addOption(
				OptionBuilder.withLongOpt("file")
				.withDescription("File containing the public key of the receiver of the delegation")
				.withArgName("Certfile")
				.hasArg()
				.isRequired(false)
				.create("f")
				);
		getOptions().addOption(
				OptionBuilder.withLongOpt("subject")
				.withDescription("The distinguished name (DN) of the receiver of the delegation")
				.withArgName("SubjectDN")
				.hasArg()
				.isRequired(false)
				.create("S")
				);
		getOptions().addOption(
				OptionBuilder.withLongOpt("target")
				.withDescription("The target file for storing the assertion")
				.withArgName("Targetfile")
				.hasArg()
				.isRequired(true)
				.create("t")
				);
		getOptions().addOption(
				OptionBuilder.withLongOpt("validity")
				.withDescription("The number of days that the delegation stays valid (default: 1)")
				.withArgName("Validity_Days")
				.hasArg()
				.isRequired(false)
				.create("V")
				);
		getOptions().addOption(
				OptionBuilder.withLongOpt(OPT_SITENAME_LONG)
				.withDescription("The name of the site to delegate trust to")
				.withArgName("Site name")
				.hasArg()
				.isRequired(false)
				.create(OPT_SITENAME)
				);
	}
	
	/**
	 * only need to connect to Registry if we have a Site name
	 */
	protected boolean skipConnectingToRegistry(){
		return !getCommandLine().hasOption(OPT_SITENAME);
	}

	@Override
	public void process(){
		super.process();
		try{
			validity=Integer.parseInt(getOption("validity", "V", "1"));
			if(getCommandLine().hasOption("f")){
				String pemFile=getOption("file", "f");
				CertificateFactory cf=CertificateFactory.getInstance("X.509");
				FileInputStream inStream=new FileInputStream(pemFile);
				X509Certificate cert=(X509Certificate)cf.generateCertificate(inStream);
				subject=cert.getSubjectX500Principal().getName();
			}
			else if(getCommandLine().hasOption("S")){
				subject=getCommandLine().getOptionValue("S");
			}
			else if(getCommandLine().hasOption(OPT_SITENAME)){
				String siteName=getCommandLine().getOptionValue(OPT_SITENAME);
				findSubjectBySiteName(siteName);
			}
			else {
				printUsage();
				throw new IllegalArgumentException("Must specify one of: cert file, subject DN or site name.");
			}
			target=getOption("target", "t", "td.out");
			doIssueTD();
		}catch(Exception e){
			error("Can't issue trust delegation.", e);
			endProcessing(ERROR_CLIENT);
		}
	}

	protected void findSubjectBySiteName(String siteName){
		super.testRegistryConnection();
		TSSClient t=new TargetSystemLister(registry,securityProperties,this).getTSS(siteName);
		EndpointReferenceType epr=t.getEPR();
		verbose("Using server at "+epr.getAddress().getStringValue());
		subject=new X500Principal(WSUtilities.extractServerIDFromEPR(epr)).getName();
	}
	
	@Override
	protected boolean requireRegistry() {
		return false;
	}


	protected TrustDelegation doIssueTD()throws Exception{
		DelegationRestrictions restr=new DelegationRestrictions(new Date(),validity,10);
		TrustDelegation td=UnicoreSecurityFactory.getETDEngine().generateTD(
				securityProperties.getCredential().getSubjectName(), 
				securityProperties.getCredential().getCertificateChain(), 
				securityProperties.getCredential().getKey(), 
				subject, 
				restr);
		FileOutputStream fos=new FileOutputStream(target);
		td.getXML().save(fos);
		fos.close();
		return td;
	}

	@Override
	public String getName(){
		return "issue-delegation";
	}
	
	@Override
	public String getArgumentList(){
		return "<cert_file> <target_file> [<validity_days>] ";
	}
	
	@Override
	public String getSynopsis(){
		return "Issue a trust delegation assertion from the current user to " +
				"the subject specified in the cert_file. The assertion will be saved to target_file. " +
				"If not specified the validity is one day.";
	}
	
	@Override
	public String getDescription(){
		return "issue trust delegation assertion";
	}

	@Override
	public String getCommandGroup(){
		return "Security";
	}
}
