package de.fzj.unicore.ucc.actions;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.util.Calendar;
import java.util.Date;

import org.apache.commons.cli.OptionBuilder;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.JobDefinitionDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.ResourcesDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.ResourcesType;
import org.unigrids.x2006.x04.services.reservation.ReservationPropertiesDocument.ReservationProperties;
import org.unigrids.x2006.x04.services.tss.TargetSystemPropertiesDocument.TargetSystemProperties;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.client.ReservationClient;
import de.fzj.unicore.uas.client.TSSClient;
import de.fzj.unicore.ucc.helpers.TargetSystemLister;
import de.fzj.unicore.ucc.runner.RandomSelection;
import de.fzj.unicore.ucc.runner.Requirement;
import de.fzj.unicore.ucc.runner.RunnerException;
import de.fzj.unicore.ucc.runner.TargetSystemFinder;
import de.fzj.unicore.ucc.util.Builder;
import de.fzj.unicore.ucc.util.UnitParser;

/**
 * Reservation support
 * 
 * @author schuller
 */
public class Reservation extends ActionBase {

	/**
	 * name of the Site to use 
	 */
	protected String siteName=null;

	/**
	 * treat resources file as JSDL doc instead of the standard JSON
	 */
	protected boolean isJSDL=false;

	/**
	 * requested start time of the reservation
	 */
	protected Calendar requestedStart=null;

	/**
	 * requested start time
	 */
	public static final String OPT_STARTTIME_LONG="start";
	public static final String OPT_STARTTIME="S";

	/**
	 * list reservations
	 */
	public static final String OPT_LIST_LONG="list";
	public static final String OPT_LIST="l";

	/**
	 * delete a reservation
	 */
	public static final String OPT_DELETE_LONG="delete";
	public static final String OPT_DELETE="d";


	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_SITENAME_LONG)
				.withDescription("Site Name")
				.withArgName("Vsite")
				.hasArg()
				.isRequired(false)
				.create(OPT_SITENAME)
				);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_JSDL_LONG)
				.withDescription("Treat resources file as JSDL document")
				.isRequired(false)
				.create(OPT_JSDL)
				);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_LIST_LONG)
				.withDescription("List your reservations.")
				.isRequired(false)
				.create(OPT_LIST)
				);


		getOptions().addOption(OptionBuilder.withLongOpt(OPT_DELETE_LONG)
				.withDescription("Delete a reservation.")
				.isRequired(false)
				.hasArg()
				.withArgName("ReservationReference")
				.create(OPT_DELETE)
				);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_STARTTIME_LONG)
				.withDescription("Requested start time for the reservation")
				.isRequired(false)
				.withArgName("Starttime")
				.hasArg()
				.create(OPT_STARTTIME)
				);
	}

	@Override
	public void process(){
		super.process();

		siteName=getCommandLine().getOptionValue(OPT_SITENAME);

		if(getBooleanOption(OPT_LIST_LONG, OPT_LIST)){
			try{
				doList();
				return;
			}catch(Exception e){
				error("Can't list reservations.", e);
				endProcessing(ERROR);
			}
		}

		if(getBooleanOption(OPT_DELETE_LONG, OPT_DELETE)){
			try{
				doDelete();
				return;
			}catch(Exception e){
				error("Can't delete reservation.", e);
				endProcessing(ERROR);
			}
		}

		isJSDL=getBooleanOption(OPT_JSDL_LONG, OPT_JSDL);
		verbose("Assuming JSDL resources file = "+isJSDL);
		String requestedStartTime=getOption(OPT_SCHEDULED_LONG, OPT_SCHEDULED);
		if(requestedStartTime!=null){
			Date req=UnitParser.extractDateTime(requestedStartTime);
			requestedStart=Calendar.getInstance();
			requestedStart.setTime(req);
			verbose("Requested start time "+requestedStart.getTime());
		}
		else{
			error("Please provide a start time!",null);
			endProcessing(ERROR);
		}

		try{
			doCreate();
		}catch(Exception e){
			error("Can't create reservation.", e);
			endProcessing(ERROR);
		}
	}

	@Override
	public String getName(){
		return "reservation";
	}

	@Override
	public String getArgumentList(){
		return "command <options>";
	}

	@Override
	public String getSynopsis(){
		return "Create and manage your reservations. " +
				"You can make a reservation at a site specified using the '" +
				OPT_SITENAME_LONG+"' option. The resources to reserve are specified using a UCC job file. " +
				"The requested start time is given using the '"+OPT_STARTTIME_LONG+"' option. The supported " +
				"date formats are 'HH:mm', 'yyyy-MM-dd HH:mm' and ISO8601. "+
				"To list your reservations, use the '"+OPT_LIST_LONG+"' option. " +
				"Use the '"+OPT_DELETE_LONG+"' option to delete a reservation, specifying the reservation " +
				"reference on the command line.";
	}

	@Override
	public String getDescription(){
		return "create and manage reservations";
	}

	@Override
	public String getCommandGroup(){
		return "Job execution";
	}

	/**
	 * find a suitable TSS for job submission.
	 * @throws RunnerException if no matching TSS can be found
	 */
	protected TSSClient findTSS(Builder builder)throws RunnerException{
		TargetSystemFinder finder=new TargetSystemFinder(registry,securityProperties,new RandomSelection(),siteName,this);
		finder.setCheckResources(true);
		builder.getRequirements().add(new ReservationRequirement());
		return finder.find(builder.getRequirements());
	}


	protected void doCreate() throws Exception{
		Builder builder=initBuilder();
		TSSClient tss=findTSS(builder);
		ResourcesType res=builder.getJob().getJobDefinition().getJobDescription().getResources();
		ResourcesDocument resources=ResourcesDocument.Factory.newInstance();
		resources.setResources(res);
		ReservationClient rrClient=tss.createReservationClient(resources, requestedStart);
		showReservationDetails(rrClient);
	}

	protected void doList()throws Exception{
		TargetSystemLister tsl=new TargetSystemLister(registry, securityProperties, this);
		lastNumberOfReservations=0;
		if(siteName!=null){
			TSSClient tss=tsl.getTSS(siteName);
			if(tss!=null){
				verbose("Listing reservations on site '"+siteName+"'");
				listReservationsOnTSS(tss);
			}
			else{
				message("Target system at site '"+siteName+"' does not exist! Check the site name or try 'ucc connect' first.");
			}
		}
		else{
			for(TSSClient tss: tsl){
				listReservationsOnTSS(tss);
			}
		}
	}

	protected void listReservationsOnTSS(TSSClient tss)throws Exception{
		if(!tss.supportsReservation()){
			message("TSS "+tss.getTargetSystemName()+" does not support reservations.");
			return;
		}
		message("Site: "+tss.getTargetSystemName());
		message(tss.getUrl());
		for(EndpointReferenceType epr: tss.getReservations()){
			showReservationDetails(new ReservationClient(epr, tss.getSecurityConfiguration()));
			lastNumberOfReservations++;
		}
	}

	protected void showReservationDetails(ReservationClient r)throws Exception{
		ReservationProperties rp=r.getResourcePropertiesDocument().getReservationProperties();
		message(" Reservation: "+rp.getReservationReference());
		message("   URL      : "+r.getUrl());
		Calendar start=rp.getStartTime();
		message("   Starting : "+(start!=null ? start : "n/a"));
		message("   Status   : "+rp.getReservationStatus());
		String desc=rp.getReservationStatusDescription();
		message("   Descript.: "+(desc!=null?desc:"n/a"));
		lastReservationID=rp.getReservationReference();
		lastReservationURL=r.getUrl();
	}

	protected void doDelete()throws Exception{
		String resID=null;
		resID=getOption(OPT_DELETE_LONG, OPT_DELETE);

		if(resID==null)throw new IllegalArgumentException("Please specify a reservation ID to delete!");

		TargetSystemLister tsl=new TargetSystemLister(registry, securityProperties, this);
		if(siteName!=null){
			TSSClient tss=tsl.getTSS(siteName);
			if(tss!=null){
				findAndDelete(resID,tss);
			}
			else{
				message("Target system at site '"+siteName+"' does not exist! Check the site name or try 'ucc connect' first.");
			}
			
		}
		else{
			for(TSSClient tss: tsl){
				findAndDelete(resID,tss);
			}
		}
	}

	protected boolean findAndDelete(String resID, TSSClient tss)throws Exception{
		boolean found=false;
		if(!tss.supportsReservation()){
			verbose("TSS "+tss.getTargetSystemName()+" does not support reservations.");
		}
		else{
			for(EndpointReferenceType epr: tss.getReservations()){
				ReservationClient r=new ReservationClient(epr, tss.getSecurityConfiguration());
				if(resID.equals(r.getReservationReference())){
					r.destroy();
					message("Deleting reservation "+resID);
					found=true;
					break;
				}
			}
		}
		return found;
	}

	public static class ReservationRequirement implements Requirement{

		@Override
		public boolean isFulfilled(TargetSystemProperties tss) {
			return tss.getSupportsReservation();
		}

		@Override
		public String getDescription() {
			return "Resource reservation support";
		}

		@Override
		public int hashCode() {
			return 1;
		}

		@Override
		public boolean equals(Object obj) {
			return obj!=null && obj instanceof ReservationRequirement;
		}

	}

	protected Builder initBuilder(){
		String jobFile=null;
		Builder builder=null;
		if(getCommandLine().getArgs().length>1){
			jobFile=getCommandLine().getArgs()[1];
			if(isJSDL){
				try{
					verbose("Loading JSDL file.");
					builder=new Builder();
					builder.setMessageWriter(this);
					builder.setRegistry(registry);
					builder.setSecurityProperties(securityProperties);
					JobDefinitionDocument jdd=JobDefinitionDocument.Factory.parse(new File(jobFile));
					builder.setJob(jdd);
				}catch(Exception e){
					error("Can't parse jsdl file <"+jobFile+">",e);
					endProcessing(ERROR_CLIENT);
				}
			}
			else {
				try{
					builder=new Builder(new File(jobFile));
					builder.setMessageWriter(this);
					builder.setRegistry(registry);
					builder.setSecurityProperties(securityProperties);
					verbose("Read job from <"+jobFile+">");
				}catch(Exception e){
					error("Can't parse job file <"+jobFile+">",e);
					endProcessing(ERROR_CLIENT);
				}
			}
		}
		else{
			try{
				message("Reading job from stdin:");
				message("");
				ByteArrayOutputStream bos=new ByteArrayOutputStream();
				int b=0;
				while((b=System.in.read())!=-1){
					bos.write(b);
				}
				builder=new Builder(bos.toString());
			}catch(Exception e){
				error("Can't read job from stdin.",e);
				endProcessing(ERROR_CLIENT);
			}
		}
		return builder;
	}

	//for unit testing
	static String lastReservationID=null;
	static String lastReservationURL=null;
	static Integer lastNumberOfReservations=null;

}
