/*
 * Copyright (c) 2011 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Sep 5, 2011
 * Author: K. Stasiak <karol.m.stasiak@gmail.com>
 */

package de.fzj.unicore.ucc.actions;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.Charset;
import org.apache.commons.cli.OptionBuilder;
import org.apache.xmlbeans.XmlOptions;

/**
 * Saves an attribute assertion to a file.
 * @author Karol Stasiak
 */
public class SaveAttributeAssertion extends ActionBase{

	@Override
	public String getName() {
		return "save-attributes";
	}


	@Override
	public String getDescription() {
		return "save attribute assertion to XML file.";
	}

	@Override
	public String getCommandGroup(){
		return "Security";
	}

	@Override
	public String getSynopsis() {
		return "Save attribute assertion to XML file. "
				+ "If given VO URL with the option -"+OPT_VO+", it connects to it and fetches an assertion. "
				+ "If given a file with an assertion with the option -"+OPT_ATTRIBUTEASSERTION+", it loads the assertion from the file. "
				+ "The attributes are filtered. "
				+ "If the option -"+OPT_ATTRIBUTELISTOUTFILE+" <file> is specified, then the assertion is "
				+ "written to the file.";
	}
	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_ATTRIBUTELISTOUTFILE_LONG)
				.withDescription("File to save assertion; empty for stdout")
				.withArgName("OutputFile")
				.hasArg()
				.isRequired(false)
				.create(OPT_ATTRIBUTELISTOUTFILE)
				);
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_ASSERTIONPRETTYPRINT_LONG)
			.withDescription("Reformat the assertion before being outputted so it is human readable. " +
					"WARNING: such reformatted assertion WILL NOT be useful for authorization.")
			.isRequired(false)
			.create(OPT_ASSERTIONPRETTYPRINT)
			);
	}

	@Override
	protected boolean skipConnectingToRegistry(){
		return
			null != getOption(OPT_ATTRIBUTEASSERTION_LONG, OPT_ATTRIBUTEASSERTION)
			 ||
			! "auto".equals(getOption(OPT_VO_LONG, OPT_VO));
	}

	@Override
	public void process(){
		super.process();
		if(!useSAMLPush){
			message("No assertion specified. Specify a VO server or a file containing an assertion.");
			endProcessing(ERROR);
		}
		if(null == getOption(OPT_ATTRIBUTEASSERTION_LONG, OPT_ATTRIBUTEASSERTION)
				&& null==getOption(OPT_VO_LONG, OPT_VO)){
			message("Neither VO server URL nor attribute assertion file to parse was provided");
			endProcessing(ERROR);
		}
		String outputFilename = getOption(OPT_ATTRIBUTELISTOUTFILE_LONG, OPT_ATTRIBUTELISTOUTFILE);
		Writer outputWriter = null;
		if(outputFilename!=null && !outputFilename.isEmpty()){
			File file = new File(outputFilename);
			try{
				outputWriter = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), Charset.forName("UTF-8")));
				writeAssertion(outputWriter);
				String assertionSource = "";
				if (voURL==null || voURL.isEmpty()){
					assertionSource = getOption(OPT_ATTRIBUTEASSERTION_LONG, OPT_ATTRIBUTEASSERTION);
				}
				else{
					assertionSource = voURL;
				}
				message("Attribute assertion from "+assertionSource+" successfully saved in "+outputFilename);
			}
			catch(IOException e){
				error("Error writing to file: "+outputFilename, e);
				endProcessing(ERROR);
			}
			finally{
				if(outputWriter!=null){
					try{outputWriter.close();}catch(Exception ex){/*IGNORE*/}
				}
			}
		}
		else{
			outputWriter = new OutputStreamWriter(System.out);
			try {
				writeAssertion(outputWriter);
				outputWriter.write("\n");
				outputWriter.flush();
			} catch (IOException ex) {
				throw new RuntimeException("Failed to write assertion onto the standard output.", ex);
			}
		}
	}

	private void writeAssertion(Writer writer) throws IOException{
		boolean prettify = getBooleanOption(OPT_ASSERTIONPRETTYPRINT_LONG, OPT_ASSERTIONPRETTYPRINT);
		XmlOptions saveOpts = new XmlOptions();
		if (prettify)
			saveOpts.setSavePrettyPrint();
		rawSAMLAssertion.getXML().save(writer, saveOpts);
		writer.flush();
	}
}
