/*
 * Copyright (c) 2011 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on 27-07-2011
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */
package de.fzj.unicore.ucc.actions;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.OptionBuilder;
import org.apache.xmlbeans.XmlException;
import org.oasisOpen.docs.wsrf.rp2.UpdateResourcePropertiesDocument;
import org.oasisOpen.docs.wsrf.rp2.UpdateResourcePropertiesDocument.UpdateResourceProperties;
import org.oasisOpen.docs.wsrf.rp2.UpdateType;
import org.unigrids.services.atomic.types.UmaskDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.ucc.helpers.EndProcessingException;
import de.fzj.unicore.wsrflite.xmlbeans.WSUtilities;
import de.fzj.unicore.wsrflite.xmlbeans.client.BaseWSRFClient;

/**
 * Changes or prints umask of a specified resource. Can be used with SMS and TSS instances.
 * 
 * @author K. Benedyczak
 */
public class Umask extends ActionBase {
	public static final String OPT_SET_LONG = "set";
	public static final String OPT_SET = "s";
	
	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_SET_LONG)
				.withDescription("Set a umask to the provided value.")
				.isRequired(false)
				.hasArg()
				.create(OPT_SET)
				);
	}

	@Override
	public String getName() {
		return "umask";
	}

	@Override
	public String getSynopsis() {
		return "Get or set umask of a remote storage (SMS) " +
				"or a target system service (TSS). Umask is used by the UNICORE" +
				" server side to control permissions of newly created files." +
				" Without the '" + OPT_SET_LONG + "' parameter the current umask is printed.";
	}
	
	@Override
	public String getDescription(){
		return "get or set a file creation umask";
	}
	
	@Override
	public String getArgumentList(){
		return "<Storage-URL | TargetSystem-URL>";
	}
	
	@Override
	public String getCommandGroup(){
		return "Data management";
	}
	
	
	@Override
	public void process() {
		super.process();
		String set = getOption(OPT_SET_LONG, OPT_SET);
		if (set != null)
			verbose("Will set a umask to " + set);
		else
			verbose("Will get a umask");
		
		CommandLine cmdLine = getCommandLine(); 
		if (cmdLine.getArgs().length != 2) {
			error("Wrong number of arguments", null);
			printUsage();
			endProcessing(ERROR_CLIENT);
		}
		String url = cmdLine.getArgs()[1];
		verbose("Operating on a address: " + url);
		
		
		BaseWSRFClient client = createWSRFClient(url);
		String umaskDocS;
		UmaskDocument umaskDoc;
		try {
			umaskDocS = client.getResourceProperty(UmaskDocument.type.getDocumentElementName());
		} catch(Exception ex) {
			error("Error getting WS-RP document. It seems that the selected service doesn't " +
					"support umask setting", ex);
			endProcessing(ERROR);
			return; //dummy
		}
		if (umaskDocS == null) {
			error("The selected service doesn't support umask setting", null);
			endProcessing(ERROR);
		}
		try {
			umaskDoc = UmaskDocument.Factory.parse(umaskDocS);
		} catch (XmlException e) {
			error("The received resource property has invalid format. " +
					"Perhaps the service version is incompatible with this client.", e);
			endProcessing(ERROR);
			return; //dummy
		}
		
		if (set == null)
			message(umaskDoc.getUmask());
		else {
			UpdateResourcePropertiesDocument reqDoc = UpdateResourcePropertiesDocument.Factory.newInstance();
			UpdateResourceProperties req = reqDoc.addNewUpdateResourceProperties();
			UpdateType update = req.addNewUpdate();
			umaskDoc = UmaskDocument.Factory.newInstance();
			umaskDoc.setUmask(set);
			WSUtilities.insertAny(umaskDoc, update);
			try {
				client.getRP().UpdateResourceProperties(reqDoc);
			} catch (Exception e) {
				error("Error updateing umask WSRP.", e);
				endProcessing(ERROR);
			}
		}
	}

	
	private BaseWSRFClient createWSRFClient(String url) {
		EndpointReferenceType e = EndpointReferenceType.Factory.newInstance();
		e.addNewAddress().setStringValue(url);
		try {
			String dn = findServerName(e);
			if (dn != null)
				WSUtilities.addServerIdentity(e, dn);
			
			BaseWSRFClient client = new BaseWSRFClient(url, e, 
					securityProperties);
			return client;
		} catch (Exception ee) {
			error("Can't create client for " + url, ee);
			endProcessing(ERROR);
			//dummy
			throw new EndProcessingException(ERROR);
		}
	}
}





