package de.fzj.unicore.ucc.authn;

import java.io.IOException;
import java.lang.reflect.Field;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;

import de.fzj.unicore.ucc.Constants;
import eu.unicore.security.canl.CachingPasswordCallback;
import eu.unicore.security.canl.CredentialProperties;
import eu.unicore.security.canl.PasswordCallback;
import eu.unicore.security.canl.TruststoreProperties;
import eu.unicore.security.xfireutil.client.XFireClientFactory;
import eu.unicore.util.configuration.PropertyMD;
import eu.unicore.util.httpclient.ClientProperties;
import eu.unicore.util.httpclient.IClientConfiguration;

/**
 * default authentication method: uses a local keystore and truststore
 * configured using the standard "credential.*" and "truststore.*" properties<br/>
 *
 * @author schuller
 */
public class KeystoreAuthN implements Authentication, Constants {

	public static final String X509="X509";
	
	@Override
	public String getName() {
		return X509;
	}

	@Override
	public String getDescription() {
		return "Uses a local keystore and optional truststore file.";
	}

	@Override
	public IClientConfiguration initSecurityProperties(Properties properties){

		ClientProperties sp=new ClientProperties(properties, getPasswordCallback(), 
				TruststoreProperties.DEFAULT_PREFIX, 
				CredentialProperties.DEFAULT_PREFIX, ClientProperties.DEFAULT_PREFIX);
		
		//message logging
		String logOut=properties.getProperty(XFireClientFactory.LOG_OUTGOING);
		if(logOut!=null)sp.getExtraSettings().setProperty(XFireClientFactory.LOG_OUTGOING, logOut);
		String logIn=properties.getProperty(XFireClientFactory.LOG_INCOMING);
		if(logIn!=null)sp.getExtraSettings().setProperty(XFireClientFactory.LOG_INCOMING, logIn);

		return sp;
	}
	
	@Override
	public void printUsage(){
		System.out.println("The following properties can be used in the UCC preference file " +
				"to configure the X509 authentication. Many of these are optional. Refer to the " +
				"manual and/or the example files.");
		System.out.println();
		System.out.println("For configuring your credential: ");
		for(Entry<String, PropertyMD> entry: getMeta(CredentialProperties.class).entrySet()){
			PropertyMD prop=entry.getValue();
			if(!prop.isHidden()){
				System.out.println(CredentialProperties.DEFAULT_PREFIX+entry.getKey()+" : "+prop.getDescription());
			}
		}
		System.out.println();
		System.out.println("For configuring your trusted CAs and certificates: ");
		for(Entry<String, PropertyMD> entry: getMeta(TruststoreProperties.class).entrySet()){
			PropertyMD prop=entry.getValue();
			if(!prop.isHidden()){
				System.out.println(TruststoreProperties.DEFAULT_PREFIX+entry.getKey()+" : "+prop.getDescription());
			}
		}
	}

	private PasswordCallback getPasswordCallback() {
		return new CachingPasswordCallback() {
			@Override
			public boolean ignoreProperties() {
				return false;
			}
			
			@Override
			public char[] getPasswordFromUser(String protectedArtifactType, String protectedArtifactDescription) {
				String r;
				System.out.println("Please enter your " + protectedArtifactType + 
						" ("+protectedArtifactDescription + ") password: ");
				try {
					r = readFromStdin();
				} catch (IOException e) {
					return null;
				}
				r.trim();
				return r.toCharArray();
			}
			
			@Override
			public boolean askForSeparateKeyPassword() {
				//TODO - or true?
				return false;
			}
		};
	}
	
	private String readFromStdin() throws IOException{
		return new jline.ConsoleReader().readLine(new Character('*'));
	}
	
	//workaround: use reflection to access possibly private Meta field
	@SuppressWarnings("unchecked")
	private Map<String,PropertyMD>getMeta(Class<?> props){
		try{
			Field f=props.getDeclaredField("META");
			f.setAccessible(true);
			return(Map<String,PropertyMD>)f.get(null);
		}catch(Exception ex){
			throw new IllegalStateException(ex);
		}
	}
	
}
