package de.fzj.unicore.ucc.helpers;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import de.fzj.unicore.ucc.InaccessibleServiceException;
import de.fzj.unicore.ucc.MessageWriter;

/**
 * wraps a Future with some additional stuff to allow retrieving information about the
 * original {@link GridTask} that was executed
 * 
 * @author schuller
 *
 * @param <T> - the type of result
 * 
 * @since 1.6.0
 */
public class GridTaskResult<T> implements Future<T> {

	private final Future<T> future;
	private final GridTaskTarget<T> target;
	
	public GridTaskResult(Future<T> future, GridTaskTarget<T> target){
		this.future=future;
		this.target=target;
	}
	
	public GridTaskTarget<T>getTarget(){
		return target;
	}
	
	@Override
	public boolean cancel(boolean mayInterruptIfRunning) {
		return future.cancel(mayInterruptIfRunning);
	}

	@Override
	public boolean isCancelled() {
		return future.isCancelled();
	}

	@Override
	public boolean isDone() {
		return future.isDone();
	}

	@Override
	public T get() throws InterruptedException, ExecutionException {
		return future.get();
	}

	@Override
	public T get(long timeout, TimeUnit unit) throws InterruptedException,
			ExecutionException, TimeoutException {
		return future.get(timeout, unit);
	}


	/**
	 * helper method to wait for some tasks to finish while imposing a timeout
	 * 
	 * @param results - list of future results
	 * @param timeout - timeout in seconds
	 * @param msg - message writer to write errors to
	 * @return list of results
	 * @throws InterruptedException
	 */
	static public <T> List<T> completeAsyncTasks(List<GridTaskResult<T>> results, int timeout, MessageWriter msg)
			throws InterruptedException{
		List<T>finalResults=new ArrayList<T>();
		for(GridTaskResult<T> f: results){
			try{
				T result=f.get(timeout, TimeUnit.SECONDS);
				if(result!=null)finalResults.add(result);
			}catch(ExecutionException ee){
				if(! (ee.getCause() instanceof InaccessibleServiceException)){
					String address=f.getTarget().getEpr().getAddress().getStringValue();
					String message="Error accessing "+address;
					msg.error(message,ee.getCause());
				}
			}catch(TimeoutException ex){
				String address=f.getTarget().getEpr().getAddress().getStringValue();
				msg.error("Timeout exceeded waiting for "+address, null);
			}
		}
		return finalResults;
	}
	
}
