package de.fzj.unicore.ucc.helpers;

import java.io.ByteArrayInputStream;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;
import net.sf.ehcache.Statistics;

/**
 * Simple cache support. Multiple caches identified by a cache ID can be used.
 * 
 * @author schuller
 */
public class ResourceCache {

	private static final ResourceCache cache = new ResourceCache();
	
	private final Map<String,Cache>caches;
	
	private final CacheManager cacheManager;
	
	private final AtomicInteger cacheHits=new AtomicInteger(0);
	
	private ResourceCache(){
		caches=new ConcurrentHashMap<String,Cache>();
		ByteArrayInputStream bis=new ByteArrayInputStream(ehCacheConfig.getBytes());
		cacheManager=CacheManager.create(bis);
	}
	
	/**
	 * config for ehcache</br>
	 * This is an in-memory cache only, 
	 * using the default LRU policy 
	 */
	private String ehCacheConfig="<ehcache>\n" +
			   "<defaultCache maxElementsInMemory=\"500\"\n"+
		        "eternal=\"false\"\n"+
		        "timeToIdleSeconds=\"300\"\n"+
		        "timeToLiveSeconds=\"300\"\n"+
		        "overflowToDisk=\"false\"\n"+
		        "diskPersistent=\"false\"\n"+
		        "diskExpiryThreadIntervalSeconds=\"120\"/>\n"+
		        "</ehcache>";
	public static ResourceCache getInstance(){
		return cache;
	}
	
	public void put(String cacheID, Object key, Object value){
		Cache c=doGetCache(cacheID);
		Element el=new Element(key,value);
		c.put(el);
	}
	
	public Object get(String cacheID, Object key){
		Cache c=doGetCache(cacheID);
		Element el=c.get(key);
		if(el==null)return null;
		cacheHits.incrementAndGet();
		return el.getObjectValue();
	}
	
	@SuppressWarnings("unchecked")
	public <T> T getAs(String cacheID, Object key, Class<T>target){
		Object o=get(cacheID,key);
		if(o==null)return null;
		return (T)o;
	}
	
	public Statistics getStatistics(String cacheID){
		Cache c=doGetCache(cacheID);
		return c.getStatistics();
	}
	

	public int getCacheHits(){
		return cacheHits.intValue();
	}
	
	private synchronized Cache doGetCache(String id){
		if(id==null)throw new NullPointerException("cache id can't be null");
		Cache c=caches.get(id);
		if(c==null){
			cacheManager.addCache(id);
			c=cacheManager.getCache(id);
			caches.put(id, c);
		}
		return c;
	}
	

	
	
}
