package de.fzj.unicore.ucc.helpers;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Logger;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.FileSystemDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.StorageManagement;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.StorageFactoryClient;
import de.fzj.unicore.uas.client.TSSClient;
import de.fzj.unicore.ucc.MessageWriter;
import de.fzj.unicore.wsrflite.xmlbeans.client.IRegistryQuery;
import eu.unicore.util.httpclient.IClientConfiguration;

/**
 * provides access to the available storages on a TSS
 * 
 * @author schuller
 */
public class StorageLister implements Iterable<StorageClient>{

	private static final Logger logger=Logger.getLogger(StorageLister.class.getName());

	final TSSClient tss;
	
	final IRegistryQuery registry;

	final IClientConfiguration securityProperties;
	
	final MessageWriter msg;
	
	public StorageLister(TSSClient tss, IClientConfiguration securityProperties, MessageWriter msg){
		this.tss=tss;
		registry=null;
		this.securityProperties=securityProperties;
		this.msg=msg;
	}
	
	public StorageLister(IRegistryQuery registry,IClientConfiguration securityProperties, MessageWriter msg){
		this.tss=null;
		this.registry=registry;
		this.securityProperties=securityProperties;
		this.msg=msg;
	}

	/**
	 * return a list of all accessible storages, 
	 * providing a pre-initialised StorageClient per instance
	 */
	public List<StorageClient> list(){
		try{
			List<EndpointReferenceType> eprs=getStorageEPRs();
			List<StorageClient> sms=new ArrayList<StorageClient>();
			for(EndpointReferenceType epr: eprs){
				sms.add(new StorageClient(epr.getAddress().getStringValue(),epr,getSecurityProperties()));
			}
		}catch(Exception e){
			logger.error("Can't create Storage Client list.",e);
		}
		return null;
	}
	
	/**
	 * return an iterator over the available storages, providing a
	 * pre-initialised StorageClient per instance
	 * 
	 * @see de.fzj.unicore.uas.client.StorageClient
	 */
	public Iterator<StorageClient> iterator(){
		try{
			final List<EndpointReferenceType> sms=getStorageEPRs();
			return new Iterator<StorageClient>(){
				Iterator<EndpointReferenceType> eprs=sms.iterator();
				public boolean hasNext() {
					return eprs.hasNext();
				}
				public StorageClient next() {
					EndpointReferenceType epr=eprs.next();
					try{
						return new StorageClient(epr.getAddress().getStringValue(),epr,getSecurityProperties());
					}catch(Exception e){
						logger.error("Can't create Storage Client.",e);
						return null;
					}
				}

				public void remove() {
					eprs.remove();
				}
				
			};
		}catch(Exception e){
			msg.error("Error listing storages",e);
			List<StorageClient>x=Collections.emptyList();
			return x.iterator();
		}
	}
	
	/**
	 * get a list of storage addresses, either from the TSSClient supplied to the constructor, or
	 * from the registry
	 */
	protected List<EndpointReferenceType>getStorageEPRs()throws Exception{
		if(tss!=null)return tss.getStorages();
		//else check storages in the registry
		List<EndpointReferenceType>result=new ArrayList<EndpointReferenceType>();
		result.addAll(registry.listAccessibleServices(StorageManagement.SMS_PORT));
		
		for(StorageFactoryClient sfc: new StorageFactoryLister(registry,securityProperties,msg)){
			msg.verbose("Storage factory at "+sfc.getEPR().getAddress().getStringValue());
			List<EndpointReferenceType>storages=sfc.getAccessibleStorages();
			if(storages!=null)result.addAll(storages);
		}
		return result;
	}
	
	protected IClientConfiguration getSecurityProperties(){
		return securityProperties;
	}
	
	/**
	 * get a Storage by name
	 * 
	 * @param storageName
	 * @return StorageClient for the storage with the given name or <code>null</code> if not found
	 */
	public StorageClient getStorage(String storageName){
		for(StorageClient c: this){
			try{
				String name=FileSystemDocument.Factory.parse(c.getResourceProperty(StorageManagement.RPFileSystem)).
					getFileSystem().getName();
				if(name.equalsIgnoreCase(storageName))return c;
			}catch(Exception e){
				
			}
		}
		return null;
	}
	
}
