package de.fzj.unicore.ucc.runner;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.unigrids.x2006.x04.services.tss.TargetSystemPropertiesDocument.TargetSystemProperties;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.client.TSFClient;
import de.fzj.unicore.uas.client.TSSClient;
import de.fzj.unicore.ucc.Constants;
import de.fzj.unicore.ucc.MessageWriter;
import de.fzj.unicore.ucc.helpers.GridTask;
import de.fzj.unicore.ucc.helpers.GridTaskResult;
import de.fzj.unicore.ucc.helpers.GridTaskTarget;
import de.fzj.unicore.ucc.helpers.TargetSystemFactoryLister;
import de.fzj.unicore.wsrflite.xmlbeans.client.IRegistryQuery;
import eu.unicore.util.Log;
import eu.unicore.util.httpclient.IClientConfiguration;

/**
 * Selects a suitable target system based on requirements
 *  
 * @author schuller
 */
public class TargetSystemFinder implements Constants{

	private final IRegistryQuery registry;
	
	private final IClientConfiguration securityProperties;

	private final MessageWriter msg;
	
	private final SiteSelectionStrategy selectionStrategy;
	
	private final String siteName;
	
	private boolean checkResources=true;
	
	private String blackList;
	
	/**
	 * @param registry
	 * @param properties
	 * @param msg
	 */
	public TargetSystemFinder(IRegistryQuery registry, 
			IClientConfiguration properties, 
			MessageWriter msg){
		this(registry,properties,null,null,msg);
	}
	
	/**
	 * @param registry
	 * @param properties
	 * @param selectionStrategy
	 * @param siteName
	 * @param msg
	 */
	public TargetSystemFinder(IRegistryQuery registry, 
			IClientConfiguration properties, 
			SiteSelectionStrategy selectionStrategy, 
			String siteName, MessageWriter msg){
		this.registry=registry;
		this.securityProperties=properties;
		this.msg=msg;
		this.selectionStrategy = selectionStrategy!=null ?
				selectionStrategy : new RandomSelection();
		this.siteName=siteName;
	}
	
	public void setCheckResources(boolean checkResources){
		this.checkResources=checkResources;
	}
	
	public void setBlackList(String blacklist){
		this.blackList=blacklist;
	}
	
	/**
	 * finds a TSS matching all requirements
	 * @param requirements
	 * @return
	 * @throws RunnerException if no match could be found
	 */
	public TSSClient find(final Collection<Requirement> requirements)throws RunnerException{
		try{
			final List<TSSClient>available=Collections.synchronizedList(new ArrayList<TSSClient>());
			TargetSystemFactoryLister tsfList=new TargetSystemFactoryLister(registry,securityProperties,msg);
			if(blackList!=null){
				msg.verbose("Using blacklist <"+blackList+">");
				String[] blacklist=blackList.trim().split(" ");
				tsfList.setBlacklist(blacklist);
			}
			
			GridTask<Boolean>task=new GridTask<Boolean>(){

				private TSFClient createClient(GridTaskTarget<Boolean> target) throws Exception {
					EndpointReferenceType epr=target.getEpr();
					return new TSFClient(epr.getAddress().getStringValue(),epr,securityProperties);
				}

				/**
				 * for the given target, get the accessible TSS instances, 
				 * and check whether they are appropriate
				 * 
				 * @param target
				 * @return </code>Boolean.TRUE</code> if a matching site was found, </code>Boolean.FALSE</code> otherwise
				 * @throws Exception
				 */
				public Boolean run(GridTaskTarget<Boolean> target)
						throws Exception {
					msg.verbose("Checking "+target.getEpr().getAddress().getStringValue());
					TSFClient tsf=createClient(target);
					List<EndpointReferenceType>tssEPRs=tsf.getAccessibleTargetSystems();
					ErrorHolder err=new ErrorHolder();
					//check site name first
					if(siteName!=null){
						String currentSite=tsf.getResourcePropertiesDocument().getTargetSystemFactoryProperties().getName();
						if(siteName.equals(currentSite)){
							if(tssEPRs.size()>0){
								EndpointReferenceType epr=tssEPRs.get(0);
								TSSClient tss=new TSSClient(epr.getAddress().getStringValue(),epr,securityProperties);
								msg.verbose("Using requested target system <"+siteName+">");
								if(checkResources && !matches(tss,requirements,err)){
									msg.verbose(err.message);
									throw new RunnerException(err.code, err.message);
								}
								available.add(tss);
								return Boolean.TRUE;
							}
						}
					}
					else{
						for(EndpointReferenceType epr: tssEPRs){
							TSSClient tss=new TSSClient(epr.getAddress().getStringValue(),epr,securityProperties);
							if(checkResources){
								if(matches(tss,requirements,err))available.add(tss);
								else return Boolean.FALSE;
							}
							else available.add(tss);
							return Boolean.TRUE;
						}
					}
					return Boolean.FALSE;
				}
			};

			List<GridTaskResult<Boolean>> results=tsfList.runOnEachTSF(task);

			//wait for tasks to finish
			GridTaskResult.completeAsyncTasks(results, 10, msg);
			
		    TSSClient tss=null;
			if(available.size()==0){
				throw new RunnerException(Runner.ERR_NO_SITE,"No matching target system available (try 'connect' or check job requirements)");
			}
			else{
				//select one
				tss=selectionStrategy.select(available);
			}
			msg.verbose("Selected TSS at "+tss.getEPR().getAddress().getStringValue());
			return tss;
		}
		catch(Exception e){
			throw new RunnerException(Runner.ERR_NO_SITE,"Can't find a target system.",e);
		}
	}
	

	/**
	 * check resource requirements
	 */
	protected boolean matches(TSSClient tssClient, Collection<Requirement> requirements, ErrorHolder error){
		try{
			if(requirements==null || requirements.size()==0){
				return true;
			}
			TargetSystemProperties props=tssClient.getResourcePropertiesDocument().getTargetSystemProperties();
			for(Requirement r: requirements){
				msg.verbose("Check requirement: "+r.getDescription());
				if(r.isFulfilled(props))continue;
				error.code=Runner.ERR_UNMET_REQUIREMENTS;
				error.message="Requirement <"+r.getDescription()+"> not fulfilled on "+tssClient.getTargetSystemName();
				return false;
			}
			return true;
		}catch(Exception e){
			error.code=Runner.ERR_SITE_UNAVAILABLE;
			error.message=Log.createFaultMessage("Can't contact target system", e);
			return false;
		}
	}
	
	static class ErrorHolder {
		String code, message;
	}

}
