package de.fzj.unicore.ucc.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.XmlOptions;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.ApplicationDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.ApplicationType;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.CreationFlagEnumeration;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.DataStagingDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.DataStagingType;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.JobDefinitionDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.JobDescriptionDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.JobDescriptionType;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.OperatingSystemTypeEnumeration;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.ResourcesDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.ResourcesType;
import org.ggf.schemas.jsdl.x2005.x11.jsdlPosix.ArgumentType;
import org.ggf.schemas.jsdl.x2005.x11.jsdlPosix.EnvironmentType;
import org.ggf.schemas.jsdl.x2005.x11.jsdlPosix.POSIXApplicationDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdlPosix.POSIXApplicationType;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.unigrids.services.atomic.types.ProtocolType;

import de.fzj.unicore.ucc.MessageWriter;
import de.fzj.unicore.ucc.helpers.DefaultMessageWriter;
import de.fzj.unicore.ucc.helpers.Location;
import de.fzj.unicore.ucc.runner.ApplicationRequirement;
import de.fzj.unicore.ucc.runner.ExecutionEnvironmentRequirement;
import de.fzj.unicore.ucc.runner.OSRequirement;
import de.fzj.unicore.ucc.runner.Requirement;
import de.fzj.unicore.wsrflite.xmlbeans.WSUtilities;
import de.fzj.unicore.wsrflite.xmlbeans.client.IRegistryQuery;
import eu.unicore.jsdl.extensions.ArgumentDocument.Argument;
import eu.unicore.jsdl.extensions.ExecutionEnvironmentDocument;
import eu.unicore.jsdl.extensions.ExecutionEnvironmentDocument.ExecutionEnvironment;
import eu.unicore.jsdl.extensions.IgnoreFailureDocument;
import eu.unicore.jsdl.extensions.ResourceRequestDocument;
import eu.unicore.jsdl.extensions.UserCmdType;
import eu.unicore.util.httpclient.IClientConfiguration;

/**
 * helper to build a Job including JSDL JobDefinition, 
 * local imports / exports, etc.
 *
 * This offers two modes of use: 1) provide the JSON directly, or
 * 2) use the API to construct the wanted job.
 * 
 * @author schuller
 */
public class Builder {

	private static final Logger logger=Logger.getLogger(Builder.class.getName());

	private List<FileUploader> imports;
	private List<FileDownloader> exports;
	private JobDefinitionDocument job;
	private JSONObject json;

	private MessageWriter msg=new DefaultMessageWriter();
	private IRegistryQuery registry=null;
	private IClientConfiguration security=null;

	private String jsonString;

	private final Set<Requirement>requirements=new HashSet<Requirement>();

	private final File sourceFile;

	private boolean initialised;

	private boolean checkLocalFiles=true;

	/**
	 * reads a JSON string from the supplied File
	 * and creates the builder from it
	 * 
	 * @param jsonFile
	 * @throws Exception
	 */
	public Builder(File jsonFile)throws Exception{
		sourceFile=jsonFile;
	}

	/**
	 * and creates the builder from the supplied JSON string
	 * 
	 * @param jsonString
	 * @throws Exception
	 */
	public Builder(String jsonString)throws Exception{
		this.jsonString=jsonString;
		sourceFile=null;
	}

	public void setRegistry(IRegistryQuery registry){
		this.registry=registry;
	}

	public void setSecurityProperties(IClientConfiguration sec){
		this.security=sec;
	}

	public void setMessageWriter(MessageWriter msg){
		this.msg=msg;
	}

	/**
	 * creates an empty builder
	 * @throws Exception
	 */
	public Builder()throws Exception{
		sourceFile=null;
		exports=new ArrayList<FileDownloader>();
		imports=new ArrayList<FileUploader>();
		jsonString="{}";
		json=new JSONObject(jsonString);
		json.put("Output",".");
		initialised=true;
	}

	private void readJsonString()throws IOException{
		FileInputStream fis=null;
		try{
			fis=new FileInputStream(sourceFile.getAbsolutePath());
			BufferedReader r=new BufferedReader(new InputStreamReader(fis));
			String l=null;
			StringBuilder b=new StringBuilder();
			while( (l=r.readLine()) !=null){
				b.append(l+"\n");
			}
			jsonString=b.toString();
		}
		finally{
			if(fis!=null)fis.close();	
		}

	}

	private synchronized void build(){
		if(initialised)return;
		initialised=true;
		try{
			if(sourceFile!=null){
				readJsonString();
			}
			json=new JSONObject(jsonString);

		}catch(IOException ex){
			msg.error("Error building job", ex);
			throw new RuntimeException(ex);
		}catch(JSONException ex){
			String message=JSONUtil.makeParseErrorMessage(jsonString, ex);
			msg.error(message, ex);
			throw new RuntimeException(ex);
		}

		try{
			imports=getLocalImports(json.optJSONArray("Imports"));
			exports=getLocalExports(json.optJSONArray("Exports"));
		}catch(Exception ex){
			msg.error("Error building imports/exports for job", ex);
			throw new RuntimeException(ex);
		}

		try{
			extractRequirements();
		}catch(Exception ex){
			msg.error("Error building job requirements", ex);
			throw new RuntimeException(ex);
		}
	}

	private void extractRequirements(){
		if(job!=null){
			extractFromJSDL();
			return;
		}

		String appName=JSONUtil.getString(json,"ApplicationName");
		String appVersion=JSONUtil.getString(json,"ApplicationVersion");
		if(appName!=null){
			ApplicationRequirement appRequired=new ApplicationRequirement(appName,appVersion);
			requirements.add(appRequired);
		}

		String os=JSONUtil.getString(json,"Operating system");
		if(os!=null){
			OperatingSystemTypeEnumeration.Enum osType=OperatingSystemTypeEnumeration.Enum.forString(os);
			if(osType!=null){
				requirements.add(new OSRequirement(osType));
			}else{
				msg.error("Operating system "+os+" not recognized (not defined in JSDL).", null);
			}
		}

		JSONObject ee=json.optJSONObject("Execution environment");
		if(ee!=null){
			String name=ee.optString("Name");
			String version=JSONUtil.getString(ee,"Version");
			requirements.add(new ExecutionEnvironmentRequirement(name,version));
		}

	}

	private void extractFromJSDL(){
		//Application
		try{
			String appName=job.getJobDefinition().getJobDescription().getApplication().getApplicationName();
			String appVersion=job.getJobDefinition().getJobDescription().getApplication().getApplicationVersion();
			if(appName!=null){
				ApplicationRequirement appRequired=new ApplicationRequirement(appName,appVersion);
				requirements.add(appRequired);
			}
		}
		catch(Exception e){}
		//OS
		try{
			OperatingSystemTypeEnumeration.Enum osType=job.getJobDefinition().getJobDescription().getResources().
					getOperatingSystem().getOperatingSystemType().getOperatingSystemName();
			if(osType!=null){
				requirements.add(new OSRequirement(osType));
			}
		}catch(Exception e){}

		//Others TODO
	}


	/**
	 * creates a JSDL jobdefinition from a json object
	 * @param json
	 */
	private void makeJob() throws Exception{
		JobDescriptionType jd=JobDescriptionType.Factory.newInstance();
		try{
			jd.setApplication(makeApplication(json));
		}catch(Exception e){
			msg.error("",e);
		}

		String jobName=getProperty("Name");
		if(jobName==null){
			if(getApplicationName()==null){
				jobName="UNICORE_Job";
			}
			else{
				jobName=getApplicationName();
			}
		}
		jd.addNewJobIdentification().setJobName(jobName);

		String jobProject=getProperty("Project");
		if(jobProject!=null){
			jd.getJobIdentification().addJobProject(jobProject);
		}

		String jobDescription=getProperty("Description");
		if(jobDescription!=null){
			if(jd.getJobIdentification()==null)jd.addNewJobIdentification();
			jd.getJobIdentification().setDescription(jobDescription);
		}
		String email=getProperty("User email");
		if(email!=null){
			if(jd.getJobIdentification()==null)jd.addNewJobIdentification();
			jd.getJobIdentification().addJobAnnotation("User email: "+email);
		}

		//"not before" for scheduling server-side processing
		try{
			String notBefore=json.getString("Not before");
			if(jd.getJobIdentification()==null)jd.addNewJobIdentification();
			jd.getJobIdentification().addJobAnnotation("notBefore: "+notBefore);
		}catch(Exception e){}

		//other annotations
		try{
			JSONArray a=json.getJSONArray("Tags");
			if(jd.getJobIdentification()==null)jd.addNewJobIdentification();
			for(int i=0; i<a.length();i++){
				jd.getJobIdentification().addJobAnnotation(a.getString(i));
			}
		}catch(Exception e){}

		try{
			jd.setResources(makeResources((JSONObject)json.get("Resources")));
		}catch(Exception e){}
		
		
		List<DataStagingType>staging=new ArrayList<DataStagingType>();
		addStageIn(staging,json.optJSONArray("Imports"));
		addStageIn(staging,json.optJSONArray("Stage in"));
		
		addStageOut(staging,json.optJSONArray("Exports"));
		addStageOut(staging,json.optJSONArray("Stage out"));
		
		jd.setDataStagingArray(staging.toArray(new DataStagingType[staging.size()]));
		
		try{
			jd=insertExecutionEnvironment(jd,json.optJSONObject("Execution environment"));
		}catch(Exception e){
			throw new IllegalArgumentException("Error adding execution environment",e);
		}

		//build the final JSDL document
		job=JobDefinitionDocument.Factory.newInstance();
		job.addNewJobDefinition().setJobDescription(jd);
	}

	private void addStageIn(List<DataStagingType>staging, JSONArray j) throws IllegalArgumentException{
		if(j==null)return;
		for(int i=0; i<j.length();i++){
			try{
				JSONObject jObj=j.getJSONObject(i);
				String target=JSONUtil.getString(jObj,"To");
				String source=JSONUtil.getString(jObj, "From");
				if(source==null)source=JSONUtil.getString(jObj, "File");//backwards compatibility
				Location l=new Location(source,registry,security,msg);

				//skip local source
				if(l.isLocal())continue;

				//try to resolve a unicore6:// URI
				if(registry!=null){
					String resolved=l.resolve();
					source=resolved;
				}

				DataStagingDocument dsd=DataStagingDocument.Factory.newInstance();
				DataStagingType d=dsd.addNewDataStaging();

				d.setFileName(target);
				d.addNewSource().setURI(source);
				
				insertCommonDataStagingOptions(dsd, jObj);

				staging.add(d);
			
			}
			catch(IllegalArgumentException iae){
				throw iae;
			}catch(Exception e){
				throw new IllegalArgumentException("Stage-in specification invalid. Syntax: \"From: <location>, To: <uspacefile>, Mode: overwrite|append|nooverwrite\"",e);
			}
		}
	}

	private void addStageOut(List<DataStagingType>staging,JSONArray j)throws IllegalArgumentException{
		if(j==null)return;
		for(int i=0;i<j.length();i++){
			try{
				JSONObject jObj=j.getJSONObject(i);
				String source=JSONUtil.getString(jObj,"From");
				if(source==null)source=JSONUtil.getString(jObj,"File");//backwards compat
				String target=jObj.getString("To");

				Location l=new Location(target,registry,security,msg);

				//skip local target
				if(l.isLocal())continue;

				//try to resolve a unicore6:// URI
				if(registry!=null){
					String resolved=l.resolve();
					target=resolved;
				}

				DataStagingDocument dsd=DataStagingDocument.Factory.newInstance();
				DataStagingType d=dsd.addNewDataStaging();
				
				d.setFileName(source);
				d.addNewTarget().setURI(target);
				
				insertCommonDataStagingOptions(dsd, jObj);
				
				staging.add(d);

			}catch(Exception e){
				throw new IllegalArgumentException("Stage-out specification invalid. Syntax: \"From: <uspacefile>, To: <location>, Mode: overwrite|append|nooverwrite\"",e);
			}
		}
	}
	
	/**
	 * adds creation mode, file system, credentials and ignoreFailure
	 * 
	 * @param dsd - data staging
	 * @param spec - the JSON object descriibung the transfer
	 */
	private void insertCommonDataStagingOptions(DataStagingDocument dsd, JSONObject spec){
		DataStagingType d=dsd.getDataStaging();
		
		//creation mode
		String creation=JSONUtil.getString(spec,"Mode","overwrite");

		if("append".equalsIgnoreCase(creation)){
			d.setCreationFlag(CreationFlagEnumeration.APPEND);
		}
		else if("nooverwrite".equalsIgnoreCase(creation)){
			d.setCreationFlag(CreationFlagEnumeration.DONT_OVERWRITE);
		}
		else d.setCreationFlag(CreationFlagEnumeration.OVERWRITE);

		//file system
		String fs=spec.optString("Filesystem", null);
		if(fs!=null){
			d.setFilesystemName(fs);
			//TODO add requirement
		}
		
		//credentials
		JSONObject cred=spec.optJSONObject("Credentials");
		if(cred!=null){
			try{
				WSUtilities.append(makeCredentials(cred), d);
			}
			catch(Exception e){
				throw new IllegalArgumentException("Can't parse Credentials structure", e);
			}
		}
		
		//ignore failure
		Boolean failOnError=Boolean.parseBoolean(JSONUtil.getString(spec,"FailOnError","true"));
		if(!failOnError){
			IgnoreFailureDocument ifd=IgnoreFailureDocument.Factory.newInstance();
			ifd.setIgnoreFailure(true);
			WSUtilities.append(ifd, dsd);
		}
	}
	
	private final static String credentialTemplate="<ac:Credential xmlns:ac=\"http://schemas.ogf.org/hpcp/2007/11/ac\">"+
	"<wsse:UsernameToken xmlns:wsse=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd\">"+
	"<wsse:Username>%1s</wsse:Username>\n"+
	"<wsse:Password>%2s</wsse:Password>\n"+
    "</wsse:UsernameToken>\n"+
    "</ac:Credential>";
	
	/**
	 * create a GFD.135 compliant Credential XML structure
	 * @param creds
	 * @return Credential XmlObject
	 * @throws IllegalArgumentException
	 */
	public XmlObject makeCredentials(JSONObject creds)throws Exception{
		String user=creds.optString("Username",null);
		if(user==null)return null;
		String pass=creds.optString("Password",null);
		return XmlObject.Factory.parse(String.format(credentialTemplate, user, pass));
	}
	

	@SuppressWarnings("unchecked")
	private ResourcesType makeResources(JSONObject j)throws Exception{
		ResourcesDocument rd=ResourcesDocument.Factory.newInstance();
		ResourcesType rt=rd.addNewResources();
		Iterator<String> it=j.keys();

		while(it.hasNext()){
			String resource=it.next();
			if("Reservation".equals(resource)){
				try{
					String reservationID=j.getString("Reservation");
					insertReservationID(reservationID, rd);
				}catch(Exception e){
					msg.error("Error processing reservation id",e);
				}
			}
			else if("Operating system".equals(resource)){
				String os=JSONUtil.getString(j,"Operating system");
				if(os!=null){
					OperatingSystemTypeEnumeration.Enum osType=getOSType(os);
					if(osType!=null){
						rt.addNewOperatingSystem().addNewOperatingSystemType().setOperatingSystemName(osType);
						requirements.add(new OSRequirement(osType));
					}else{
						msg.error("Operating system "+os+" not recognized (not defined in JSDL).", null);
					}
				}
			}
			else if("Runtime".equals(resource)){
				String runtime = j.getString("Runtime");
				if(runtime!=null){
					runtime=String.valueOf(UnitParser.getTimeParser(1).getDoubleValue(runtime));
					rt.addNewIndividualCPUTime().addNewExact().setStringValue(runtime);
				}
			}
			else if("Memory".equals(resource)){
				String memory = j.getString("Memory");
				if(memory!=null){
					memory=String.valueOf(UnitParser.getCapacitiesParser(1).getDoubleValue(memory));
					rt.addNewIndividualPhysicalMemory().addNewExact().setStringValue(memory);
				}
			}
			else if("CPUs".equals(resource)){
				String totalCPUs = j.getString("CPUs");
				if(totalCPUs!=null){
					rt.addNewTotalCPUCount().addNewExact().setStringValue(totalCPUs);
				}
			}
			else if("Nodes".equals(resource)){
				String nodes = j.getString("Nodes");
				if(nodes!=null){
					rt.addNewTotalResourceCount().addNewExact().setStringValue(nodes);

				}
			}
			else if("CPUsPerNode".equals(resource)){
				String cpus = j.getString("CPUsPerNode");
				if(cpus!=null){
					rt.addNewIndividualCPUCount().addNewExact().setStringValue(cpus);
				}
			}
			else{
				//generic resource
				try{
					String req=j.getString(resource);
					insertResourceRequest(resource,req,rd);
				}catch(Exception e){
					msg.error("Error processing resource request for <"+resource+">",e);
				}
			}
		}

		return rt;
	}

	private void insertResourceRequest(String name, String value, ResourcesDocument target)throws Exception{
		ResourceRequestDocument rrd=ResourceRequestDocument.Factory.newInstance();
		rrd.addNewResourceRequest().setName(name);
		rrd.getResourceRequest().setValue(value);
		WSUtilities.append(rrd, target);
	}

	private void insertReservationID(String id, ResourcesDocument target)throws Exception{
		String resID="<u6rr:ReservationReference xmlns:u6rr=\"http://www.unicore.eu/unicore/xnjs\">"+id+"</u6rr:ReservationReference>";
		XmlObject o=XmlObject.Factory.parse(resID);
		//and append to resources doc...
		WSUtilities.append(o, target);
	}

	@SuppressWarnings("unchecked")
	private JobDescriptionType insertExecutionEnvironment(JobDescriptionType job, JSONObject jsonEE)throws Exception{
		if(jsonEE==null)return job;
		ExecutionEnvironmentDocument eed=ExecutionEnvironmentDocument.Factory.newInstance();
		ExecutionEnvironment ee=eed.addNewExecutionEnvironment();
		String name=(String)jsonEE.get("Name");
		ee.setName(name);

		JSONObject args=jsonEE.optJSONObject("Arguments");
		if(args!=null){
			Iterator<String> it=args.keys();
			while(it.hasNext()){
				name=it.next();
				String value=args.getString(name);
				Argument a=ee.addNewArgument();
				a.setName(name);
				a.setValue(value);
			}
		}

		JSONArray opts=jsonEE.optJSONArray("Options");
		if(opts!=null){
			for(int i=0;i<opts.length();i++){
				ee.addNewOption().setName(opts.getString(i));
			}
		}

		JSONArray pre=jsonEE.optJSONArray("Precommands");
		if(pre!=null){
			for(int i=0;i<pre.length();i++){
				ee.addNewPreCommand().setName(pre.getString(i));
			}
		}
		JSONArray post=jsonEE.optJSONArray("Postcommands");
		if(post!=null){
			for(int i=0;i<post.length();i++){
				ee.addNewPostCommand().setName(post.getString(i));
			}
		}
		String userPre=JSONUtil.getString(jsonEE,"User precommand");
		if(userPre!=null){
			UserCmdType uct=UserCmdType.Factory.newInstance();
			uct.setStringValue(userPre); //TODO control whether to run on login node
			ee.setUserPreCommand(uct);
		}

		String userPost=JSONUtil.getString(jsonEE,"User postcommand");
		if(userPost!=null){
			UserCmdType uct=UserCmdType.Factory.newInstance();
			uct.setStringValue(userPost); //TODO control whether to run on login node
			ee.setUserPostCommand(uct);
		}

		JobDescriptionDocument d=JobDescriptionDocument.Factory.newInstance();
		d.setJobDescription(job);
		WSUtilities.append(eed, d);
		return d.getJobDescription();
	}

	@SuppressWarnings("unchecked")
	private ApplicationType makeApplication(JSONObject j)throws Exception{
		ApplicationDocument ad=ApplicationDocument.Factory.newInstance();
		ApplicationType app=ad.addNewApplication();

		String val2=JSONUtil.getString(j,"ApplicationVersion");
		if(val2!=null)app.setApplicationVersion(val2);

		String val=JSONUtil.getString(j,"ApplicationName");
		if(val!=null){
			app.setApplicationName(val);
			ApplicationRequirement appRequired=new ApplicationRequirement(val,val2);
			requirements.add(appRequired);
		}


		POSIXApplicationDocument pd=POSIXApplicationDocument.Factory.newInstance();
		POSIXApplicationType p=pd.addNewPOSIXApplication();

		val=JSONUtil.getString(j,"Executable");
		if(val!=null)p.addNewExecutable().setStringValue(val);

		JSONArray array=j.optJSONArray("Arguments");
		if(array!=null)	p.setArgumentArray(makeArgs(array));

		array=j.optJSONArray("Environment");
		if(array!=null)	p.setEnvironmentArray(makeEnvironment(array));

		JSONObject params=j.optJSONObject("Parameters");
		if(params!=null){
			//set as key=value
			Iterator<String>it=params.keys();
			while(it.hasNext()){
				String key=it.next();
				val=params.getString(key);
				EnvironmentType et=p.addNewEnvironment();
				et.setName(key);
				et.setStringValue(val);
			}
		}


		val=JSONUtil.getString(j,"User name");
		if(val!=null)p.addNewUserName().setStringValue(val);

		val=JSONUtil.getString(j,"Group");
		if(val!=null)p.addNewGroupName().setStringValue(val);

		val=JSONUtil.getString(j,"Stdin");
		if(val!=null)p.addNewInput().setStringValue(val);

		val=JSONUtil.getString(j,"Stdout");
		if(val!=null)p.addNewOutput().setStringValue(val);

		val=JSONUtil.getString(j,"Stderr");
		if(val!=null)p.addNewError().setStringValue(val);

		WSUtilities.append(pd, ad);
		return app;
	}


	private ArgumentType[] makeArgs(JSONArray j) {
		ArgumentType[] args=new ArgumentType[j.length()];
		for (int i = 0; i < args.length; i++) {
			args[i]=ArgumentType.Factory.newInstance();
			try{
				args[i].setStringValue(j.getString(i));
			}catch(JSONException ex){
				throw new IllegalArgumentException("Error parsing argument "+i+" in argument array! ",ex);
			}
		}
		return args;
	}

	private EnvironmentType[] makeEnvironment(JSONArray j){
		EnvironmentType[] args=new EnvironmentType[j.length()];
		for (int i = 0; i < args.length; i++) {
			try{
				String val=j.getString(i);
				String[] split=val.split("=",2);
				args[i]=EnvironmentType.Factory.newInstance();
				args[i].setName(split[0].trim());
				if(split.length>1)args[i].setStringValue(split[1].trim());
				else args[i].setStringValue("");
			}catch(JSONException ex){
				throw new IllegalArgumentException("Error parsing entry "+i+" in environment array! ",ex);
			}
		}
		return args;
	}

	private List<FileUploader> getLocalImports(JSONArray j)throws IllegalArgumentException, FileNotFoundException{
		List<FileUploader>result=new ArrayList<FileUploader>();
		if(j!=null){
			for (int i = 0; i < j.length(); i++) {
				String source,target;
				Mode mode=Mode.overwrite;
				boolean failOnError=true;
				try{
					JSONObject jObj=j.getJSONObject(i);
					source=JSONUtil.getString(jObj,"From");
					if(source==null)source=JSONUtil.getString(jObj,"File");//backwards compatibility
					target=JSONUtil.getString(jObj,"To");
					String creation=JSONUtil.getString(jObj,"Mode","overwrite");
					failOnError=Boolean.parseBoolean(JSONUtil.getString(jObj,"FailOnError","true"));
					mode=Mode.valueOf(creation);
				}catch(Exception e){
					throw new IllegalArgumentException("File import specification invalid. Syntax: \"From: <localfile>, To: <uspacefile>, Mode: overwrite|append|nooverwrite\"");
				}
				if(source==null || target==null){
					throw new IllegalArgumentException("File import specification invalid. Syntax: \"From: <localfile>, To: <uspacefile>, Mode: overwrite|append|nooverwrite\"");
				}
				Location l=new Location(source,registry,security,msg,false);
				if(l.isLocal()){
					failOnError=failOnError & checkLocalFiles;
					result.add(new FileUploader(source,target,mode,failOnError));
					msg.verbose("Local file import: "+source+" -> "+target);
				}
			}
		}
		return result;
	}

	private List<FileDownloader> getLocalExports(JSONArray j)throws IllegalArgumentException{
		List<FileDownloader>result=new ArrayList<FileDownloader>();
		if(j!=null){
			for (int i = 0; i < j.length(); i++) {
				String source,target;
				Mode mode=Mode.overwrite;
				boolean failOnError=true;
				try{
					JSONObject jObj=j.getJSONObject(i);
					source=JSONUtil.getString(jObj,"From");
					if(source==null)source=JSONUtil.getString(jObj,"File");//backwards compatibility
					target=jObj.getString("To");
					String creation=JSONUtil.getString(jObj,"Mode","overwrite");
					mode=Mode.valueOf(creation);
					failOnError=Boolean.parseBoolean(JSONUtil.getString(jObj,"FailOnError","true"));
				}catch(Exception e){
					throw new IllegalArgumentException("Local export specification invalid. Syntax: \"From: <uspacefile | u6://address>, To: <localfile>, Mode: overwrite|append|nooverwrite\"");
				}
				if(source==null || target==null){
					throw new IllegalArgumentException("Local export specification invalid. Syntax: \"From: <uspacefile | u6://address>, To: <localfile>, Mode: overwrite|append|nooverwrite\"");
				}
				Location l=new Location(target,registry,security,msg,false);
				if(l.isLocal()){
					result.add(new FileDownloader(source,target,mode,failOnError));
					msg.verbose("File export to client: "+source+" -> "+target);
				}
			}
		}
		return result;
	}

	//list of common OSs for which we want to ignore case
	static final String[] knownOSs=new String[]{"LINUX", "MACOS", "AIX", 
		"FreeBSD", "NetBSD", "Solaris", "WINNT", "IRIX", "HPUX", "Unknown"};
	
	OperatingSystemTypeEnumeration.Enum getOSType(String os){
		for(String o: knownOSs){
			if(o.equalsIgnoreCase(os)){
				return OperatingSystemTypeEnumeration.Enum.forString(o);
			}
		}
		return OperatingSystemTypeEnumeration.Enum.forString(os);
	}
	
	public List<FileDownloader> getExports() {
		build();
		return exports;
	}

	public List<FileUploader> getImports() {
		build();
		return imports;
	}

	/**
	 * get the job represented by this builder object
	 * 
	 * @return a JSDL job definition doc
	 */
	public JobDefinitionDocument getJob() throws Exception{
		build();
		if(job==null)makeJob();
		return job;
	}

	/**
	 * set the job 
	 */
	public void setJob(JobDefinitionDocument job) throws Exception{
		this.job=job;
	}

	public void setProperty(String key, String value) {
		build();
		try{
			json.put(key, value);
		}catch(Exception e){logger.error("",e);}
	}

	/**
	 * returns the given property, or null if not found
	 * @param key - the property key 
	 */
	public String getProperty(String key) {
		build();
		return JSONUtil.getString(json, key);
	}

	public String getProperty(String key,String defaultValue) {
		build();
		return JSONUtil.getString(json, key,defaultValue);
	}

	public void writeTo(Writer os) {
		build();
		try{
			os.write(json.toString(2)+"\n");
			os.flush();
		}catch(Exception e){logger.error("",e);}
	}

	public String toString(){
		build();
		try{
			return json.toString(2);
		}
		catch(Exception e){}
		return super.toString()+"<invalid JSON object>";
	}

	public Builder setJobName(String name){
		try{
			json.put("Name", name);
		}catch(JSONException ex){}
		return this;
	}

	public Builder setApplication(String name, String version){
		try{
			json.put("ApplicationName", name);
			if(version!=null)json.put("ApplicationVersion", version);
		}catch(JSONException ex){}
		return this;
	}

	public String getApplicationName(){
		build();
		return JSONUtil.getString(json, "ApplicationName");
	}

	public String getApplicationVersion(){
		build();
		return JSONUtil.getString(json, "ApplicationVersion");
	}

	public Builder setExecutable(String exec){
		try{
			json.put("Executable",exec);
		}catch(JSONException e){}
		return this;
	}

	public Builder withImport(String from, String to)throws FileNotFoundException{
		return withImport(from, to, Mode.overwrite, true);
	}

	public Builder withImport(String from, String to, Mode mode, boolean failOnError)throws FileNotFoundException{
		imports.add(new FileUploader(from,to,mode,failOnError));
		return this;
	}

	public Builder withExport(String from, String to){
		exports.add(new FileDownloader(from, to,Mode.overwrite));
		return this;
	}

	public Builder hasArgument(String arg){
		build();
		getArgs().put(arg);
		return this;
	}

	public Builder hasArguments(String[] args){
		build();
		for(String a: args){
			getArgs().put(a);
		}
		return this;
	}

	public Builder setEnvironment(String key, String value){
		build();
		getEnv().put(key+"="+value);
		return this;
	}

	public Builder writeOutputTo(String outputLocation){
		build();
		try{
			json.put("Output", outputLocation);
		}catch(JSONException e){}
		return this;
	}

	public Builder asUser(String userName){
		build();
		try{
			json.put("User name", userName);
		}catch(JSONException e){}
		return this;
	}

	public Builder asProject(String project){
		build();
		try{
			json.put("Project", project);
		}catch(JSONException e){}
		return this;
	}

	public Builder setStdin(String input){
		build();
		try{
			json.put("Stdin", input);
		}catch(JSONException e){}
		return this;
	}

	public Builder setStdout(String output){
		build();
		try{
			json.put("Stdout", output);
		}catch(JSONException e){}
		return this;
	}

	public Builder setStderr(String error){
		build();
		try{
			json.put("Stderr", error);
		}catch(JSONException e){}
		return this;
	}

	/**
	 * get the preferred file transfer protocols (if defined)
	 */
	public ProtocolType.Enum[] getPreferredProtocols(){
		build();
		try{
			String protocolsP=getProperty("Preferred protocols","");
			String[] protocols=protocolsP.split(" +");
			ProtocolType.Enum[] result=new ProtocolType.Enum[protocols.length];
			for(int i=0;i<result.length;i++){
				result[i]=ProtocolType.Enum.forString(protocols[i]);
			}
			return result;
		}catch(Exception e){
			msg.error("Error setting up preferred protocols", e);
			return new ProtocolType.Enum[0];
		}
	}

	public Set<Requirement>getRequirements(){
		build();
		return requirements;
	}

	/**
	 * get the site name, defined by the "Site" parameter in the job description file
	 * @return the site name or <code>null</code> if not set
	 */
	public String getSite(){
		return getProperty("Site");
	}

	/**
	 * get the requested lifetime of the job
	 * @return lifetime in seconds, or -1 if not set
	 */
	public int getLifetime(){
		build();
		String lifetime =JSONUtil.getString(json,"Lifetime");
		if(lifetime!=null && lifetime.length()>0){
			return (int)UnitParser.getTimeParser(0).getDoubleValue(lifetime);
		}
		else return -1;
	}

	public Builder setSite(String site){
		build();
		try{
			json.put("Site", site);
		}catch(JSONException e){}
		return this;
	}

	public void setCheckLocalFiles(boolean checkLocalFiles){
		this.checkLocalFiles=checkLocalFiles;
	}

	private JSONArray getArgs(){
		build();
		try{
			return (JSONArray)json.get("Arguments");
		}
		catch(JSONException ex){
			try{
				json.put("Arguments", new JSONArray());
				return (JSONArray)json.get("Arguments");
			}catch(JSONException e){
				throw new RuntimeException(e);
			}
		}
	}

	private JSONArray getEnv(){
		build();
		try{
			return (JSONArray)json.get("Environment");
		}
		catch(JSONException ex){
			try{
				json.put("Environment", new JSONArray());
				return (JSONArray)json.get("Environment");
			}catch(JSONException e){
				throw new RuntimeException(e);
			}
		}
	}

	public static String getJobSample(){
		StringBuilder s=new StringBuilder();
		writeLine("{",s);
		writeLine("# example UCC job",s);
		writeLine("#",s);
		writeLine("# either application name OR executable must be given",s);
		writeLine("# all other fields are optional",s);
		writeLine("#",s);
		writeLine("  ApplicationName: \"Date\", ApplicationVersion: \"1.0\",",s);
		writeLine("# or:  Executable: \"/bin/date\",",s);
		writeLine("  Arguments: [\"-l\", \"-v\", ],",s);
		writeLine("  Parameters: { VERBOSE: true, INPUTFILE: foo.txt},",s);
		writeLine("  Environment: [\"VERBOSE=true\", \"TIMING=false\", ],",s);
		writeLine("  Stdout: stdout, Stderr: stderr, Stdin: stdin,",s);
		writeLine("  Imports: [",s);
		writeLine("    { From: \"localFile | remoteFile\", To: \"jobdirFile\", Mode: \"overwrite | append | nooverwrite\" , FailOnError: true|false },",s);
		writeLine("#   more imports of this form can be added",s);
		writeLine("  ],",s);
		writeLine("  Exports: [",s);
		writeLine("    { From: \"jobdirFile\", To: \"localFile | remoteFile\", Mode: \"overwrite | append | nooverwrite\" }, ",s);
		writeLine("#   more exports of this form can be added",s);
		writeLine("  ],",s);
		writeLine("  Name: \"my test job\",",s);
		writeLine("  User name: remotelogin,",s);
		writeLine("  User email: foo@bar.org,",s);
		writeLine("  Group: remotegroup,",s);
		writeLine("  Project: some_project,",s);
		writeLine("  Tags: [\"testing\", \"demo job\"] ,",s);
		writeLine("  Not before: \"2011-11-11T12:00:00+200\",",s);
		writeLine("  Site: \"DEMO-SITE\",",s);
		writeLine("  Resources: {",s);
		writeLine("    CPUs: 4,",s);
		writeLine("#   alternatively: ",s);
		writeLine("    Nodes: 2,",s);
		writeLine("    CPUsPerNode: 2,",s);
		writeLine("    Memory: 2048M,",s);
		writeLine("    Runtime: 4h,",s);
		writeLine("    Operating system: LINUX,",s);
		writeLine("    Reservation: 1234,",s);
		writeLine("  },",s);
		writeLine("  Execution environment: {",s);
		writeLine("    Name: OpenMPI,",s);
		writeLine("    Arguments: {",s);
		writeLine("      Processors: 2,",s);
		writeLine("    },",s);
		writeLine("    Options: [ VERBOSE, ],",s);
		writeLine("    Precommands: [ MODULE_TRACE, ],",s);
		writeLine("    Postcommands: [ CLEANUP, ],",s);
		writeLine("    User precommand: \"touch /home/myname/somefile\" ,",s);
		writeLine("    User postcommand: \"touch /home/myname/someotherfile\" ,",s);
		writeLine("  },",s);
		writeLine("  Preferred protocols: \"BFT RBYTEIO\",",s);
		writeLine("  Output: \"localdir\",",s);
		writeLine("  Lifetime: 24h,",s);
		writeLine("  Update interval: 10,",s);
		writeLine("}",s);
		return s.toString();
	}

	/**
	 * validate the given JSDL job against the JSDL schema
	 * @param jdd - the JSDL document to validate
	 * @param msg - the {@link MessageWriter} to use, or <code>null</code> if no messages should be logged
	 * @return <code>true</code> if the document is valid, <code>false</code> otherwise
	 */
	public static boolean isValidJSDL(JobDefinitionDocument jdd, MessageWriter msg){
		List<String>validationErrors=new ArrayList<String>();
		XmlOptions opts=new XmlOptions();
		opts.setErrorListener(validationErrors);
		jdd.validate(opts);
		int numErrors=validationErrors.size();
		boolean isValid = numErrors==0;
		if(numErrors>0){
			if(msg!=null){
				msg.message("There were "+numErrors+" JSDL validation errors, check client log, or re-run in verbose mode " +
						"to see more details.");
				for(int i=0; i<numErrors; i++){
					msg.verbose(String.valueOf(validationErrors.get(i)));
				}
			}
		}
		return isValid;
	}

	static String lineSep=System.getProperty("line.separator");
	private static void writeLine(String line,StringBuilder sb){
		sb.append(line);
		sb.append(lineSep);
	}


}
