package de.fzj.unicore.ucc.cisprovider;

import java.math.BigInteger;

import org.apache.commons.cli.OptionBuilder;
import org.ogf.schemas.glue.x2009.x03.spec20R1.DomainsDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.cisprovider.CISInfoProvider;
import de.fzj.unicore.cisprovider.xmlbeans.GetGlueInfoRequestDocument;
import de.fzj.unicore.cisprovider.xmlbeans.GetGlueInfoResponseDocument;
import de.fzj.unicore.cisprovider.xmlbeans.GetInfoRequestDocument;
import de.fzj.unicore.cisprovider.xmlbeans.GetInfoResponseDocument;

import de.fzj.unicore.uas.client.BaseUASClient;
import de.fzj.unicore.uas.client.TSSClient;
import de.fzj.unicore.ucc.actions.ActionBase;
import de.fzj.unicore.ucc.helpers.TargetSystemLister;

/**
 * a UCC command to query the CIP (Common Information Provider)
 * 
 * @author schuller
 */
public class CIPQuery extends ActionBase {
	
	public static final String OPT_TOPIC_LONG="topic";
	public static final String OPT_TOPIC="t";

	public static final String OPT_MAXRECORDS_LONG="maxrecords";
	public static final String OPT_MAXRECORDS="m";

	public static final String OPT_GLUE_LONG="glue";
	public static final String OPT_GLUE="g";

	public static final String TSS = "TargetSystemService";
	
	protected String cisURL;
	
	private CISInfoProvider cis;


	protected String siteName;

	protected String topic;
	
	protected long maxRecords;
	
	protected boolean glue=false;
	
	public String getName(){
		return "cip-query";
	}
	
	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_SITENAME_LONG)
				.withDescription("Site Name")
				.withArgName("Vsite")
				.hasArg()
				.isRequired(false)
				.create(OPT_SITENAME)
				);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_TOPIC_LONG)
				.withDescription("Topic to query")
				.withArgName("Topic")
				.hasArg()
				.isRequired(false)
				.create(OPT_TOPIC)
				);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_MAXRECORDS_LONG)
				.withDescription("Maximum number of records")
				.withArgName("MaxRecords")
				.hasArg()
				.isRequired(false)
				.create(OPT_MAXRECORDS)
				);
		
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_GLUE_LONG)
				.withDescription("Get information in GLUE format")
				.isRequired(false)
				.create(OPT_GLUE)
				);
	}

	@Override
	public void process() {
		super.process();
		
		glue=getBooleanOption(OPT_GLUE_LONG, OPT_GLUE);
		
		topic = getCommandLine().getOptionValue(OPT_TOPIC);
		if(topic==null)topic="/";
		verbose("Requested topic '"+topic+"'");
		
		maxRecords=Long.parseLong(getCommandLine().getOptionValue(OPT_MAXRECORDS, "100"));
		verbose("Maximum number of records ="+maxRecords);
		
		siteName=getCommandLine().getOptionValue(OPT_SITENAME);
		if(siteName!=null){
			for(TSSClient c: new TargetSystemLister(registry,securityProperties,this)){
				String tName=c.getTargetSystemName();
				if(siteName!=null){
					if(!siteName.equals(tName))continue;
				}
				String url=c.getEPR().getAddress().getStringValue();
				verbose("Checking TargetSystem <"+tName+"> at "+url);
				buildAddressFromTSSUrl(url);
			}
		}
		else{
			//use given URL...
			if(getCommandLine().getArgs().length>1){
				cisURL=getCommandLine().getArgs()[1];
			}
		}
		if(cisURL==null){
			message("Please specify a URL or a sitename");
			printUsage();
			endProcessing(1);
		}
		verbose("Querying CIP at "+cisURL);
		EndpointReferenceType epr=EndpointReferenceType.Factory.newInstance();
		epr.addNewAddress().setStringValue(cisURL);
		try{
			cis=new BaseUASClient(cisURL,epr,securityProperties).makeProxy(CISInfoProvider.class);
			String result=glue?getGlueInfo():getInfo();
			message(result);
		}
		catch(Exception e){
			error("Error querying CIS info provider at "+cisURL,e);
		}
	}

	protected String getInfo()throws Exception{
		GetInfoRequestDocument req=GetInfoRequestDocument.Factory.newInstance();
		req.addNewGetInfoRequest();
		req.getGetInfoRequest().setTopic(topic);
		req.getGetInfoRequest().setMaxNumberOfRecords(BigInteger.valueOf(maxRecords));
		verbose("Query: "+req);
		GetInfoResponseDocument res=cis.getInfo(req);
		if(res==null)throw new Exception("Received null result. Maybe the CIP is not deployed at the site?");
		return res.toString();
	}
	
	protected String getGlueInfo()throws Exception{
		GetGlueInfoRequestDocument req=GetGlueInfoRequestDocument.Factory.newInstance();
		req.addNewGetGlueInfoRequest();
		verbose("Query: "+req);
		GetGlueInfoResponseDocument res=cis.getGlueInfo(req);
		if(res==null)throw new Exception("Received null result. Maybe the CIP is not deployed at the site?");
		DomainsDocument dd = DomainsDocument.Factory.newInstance();
		dd.addNewDomains();
		dd.setDomains(res.getGetGlueInfoResponse().getDomains());
		return dd.toString();
	}
	
	protected void buildAddressFromTSSUrl(String url){
		int i=url.indexOf(TSS);
		cisURL=url.substring(0, i)+CISInfoProvider.SERVICENAME;
	}
	
	@Override
	public String getDescription(){
		return "query a CIS Infoprovider at a UNICORE site";
	}
	
	@Override
	public String getSynopsis(){
		return "Query a CIS info provider service. " +
				"The VSite can be specified " +
				"using the '-s' option. Alternatively the " +
				"URL of the CIP service can be provided on the commandline.";
	}

	@Override
	public String getArgumentList() {
		return "[CIP-URL]";
	}

}
