package de.fzj.unicore.ucc.bes;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.List;
import java.util.Random;

import org.apache.commons.cli.OptionBuilder;
import org.apache.xmlbeans.XmlException;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.bes.client.FactoryClient;
import de.fzj.unicore.ucc.actions.ActionBase;

public class BESActionBase extends ActionBase implements BESUCCConstants {
	
	protected String siteName;
	protected FactoryClient factory;
	protected String siteUrl;
	
	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();
		getOptions().addOption(
				OptionBuilder.withLongOpt(OPT_SITENAME_LONG).withArgName(
						"BesFactory Url").withDescription("BESFactory Url")
						.hasArg().isRequired(false).create(OPT_SITENAME));

	}
	
	
	
	@Override
	public void process() {
		super.process();
		
		Boolean contactRegistry = Boolean.parseBoolean(properties.getProperty("contact-registry"));
		
		if(!contactRegistry){
			
				verbose("Bypassing registry.");
				siteName = getOption(OPT_SITENAME_LONG, OPT_SITENAME);
				
				siteUrl = null;
				
				if(siteName == null || siteName.length() < 1){
					message("Please provide the site name with -s option. examples bes.1|http://mybessite|Path to BES EPR file");
					endProcessing(1);
				}
					
				try {
					URL url1 = new URL(siteName);
					// only taking http(s) urls
					if(!siteName.startsWith("file://"))
						contactBESByUrl(siteName);
					else {
						File f = new File(url1.getPath());
						verbose("Checking file "+f.getPath());
						validateFile(f);
						contactBESByEPRFile(f);
					}
				} catch (MalformedURLException e) {
					
					if(siteName.startsWith("bes.")){
						siteUrl = properties.getProperty(siteName);
						
						if(siteUrl==null || siteUrl.length()<1){
						    message("Either BES Url is null or misspelled. Please specify a valid BES Url..");
						    endProcessing(1);
						}

						try{
							URL url = new URL(siteUrl);
							
							if(siteUrl.startsWith("file://")){
								siteUrl = url.getPath();
								throw new MalformedURLException();
							}
								
							contactBESByUrl(siteUrl);
							
						}catch(MalformedURLException e1){
							// siteurl is a file
							File f = new File(siteUrl);
							validateFile(f);
							contactBESByEPRFile(f);
							
						}
						
					}
					else{
						message("Couldn't reach the BES: '" +siteUrl+ "' URL.");
						message("Please specify a BES URL or name.");
						endProcessing(1);
					}
				}
				
			}
			else{
				super.initRegistryClient();
				try{
					findBES();
				}catch(Exception ex){
					error("Can't find a BES factory.",ex);
				}
			}
	
	}
	
	
	private void validateFile(File f){
		
		if(f.isDirectory()){
			error("BES EPR file path <"+ siteUrl +"> is a directory. Please specify a valid BES EPR file path.",null);
			endProcessing(1);
		}
		
		if(!f.canRead() || !f.exists()){
			error("BES EPR file path <"+ siteUrl +"> doesn't exist or you don't have access privileges. "+
				  "Please specify a valid EPR file path.",null);
			endProcessing(1);
		}
	}
	
	protected void contactBESByUrl(String siteUrl){
		verbose("Contacting BES via Url "+siteUrl+" ..");
		EndpointReferenceType eprt = EndpointReferenceType.Factory.newInstance();
		eprt.addNewAddress().setStringValue(siteUrl);
		try {
			factory = new FactoryClient(siteUrl, eprt, securityProperties);
		} catch (Exception e) {
			error("Couldn't contact BES <"+siteUrl+">", e);
			endProcessing(1);
		}
	}
	
	
	protected void contactBESByEPRFile(File eprFilePath){
		verbose("Contacting BES via EPR file.");
		EndpointReferenceType eprt = null;
		try {
			eprt = EndpointReferenceType.Factory.parse(eprFilePath);
		} catch (XmlException e) {
			error("Error parsing BES EPR file.", e);
			endProcessing(1);
		} catch (IOException e) {
			error("Error loading BES EPR file.", e);
			endProcessing(1);
		}
		try {
			factory = new FactoryClient(eprt.getAddress().getStringValue(), eprt, securityProperties);
		} catch (Exception e) {
			error("Couldn't contact BES <"+eprt.getAddress().getStringValue()+">", e);
			endProcessing(1);
		}
		
	}
	
	protected void findBES() throws Exception{
		List<FactoryClient> list = new BESFactoryLister(registry,securityProperties).list();
		int size = list.size();
		
		if (size > 0) {
			if (siteName != null) {
				for (FactoryClient f : list) {
					if (f.getEPR().getAddress().getStringValue().contains(siteName)) {
						factory = f;
						return;
					}
				}
			}
			else{
				factory = list.get(new Random().nextInt(size));
			}

		} else {
			message("No BESFactory instance found in the registry.");
			endProcessing(1);
		}

	}
	
	@Override
	protected void initRegistryClient() {
		// do nothing
	}
	
}
