package de.fzj.unicore.ucc.workflow;

import java.io.File;
import java.util.Iterator;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.commons.cli.OptionBuilder;
import org.chemomentum.common.impl.workflow.WorkflowManagementClient;
import org.chemomentum.common.ws.IWorkflowFactory;
import org.chemomentum.common.ws.WorkflowManagement;
import org.chemomentum.workflow.xmlbeans.JobReferenceDocument;
import org.chemomentum.workflow.xmlbeans.OutputFilesDocument;
import org.chemomentum.workflow.xmlbeans.WorkflowFactoryPropertiesDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.client.BaseUASClient;
import de.fzj.unicore.uas.client.EnumerationClient;
import de.fzj.unicore.ucc.actions.ListActionBase;

/**
 * creates a workflow listing
 * @author schuller
 */
public class WorkflowInfo extends ListActionBase<WorkflowManagementClient> {

	private boolean listFiles;

	private boolean listJobs;


	@Override
	public String getName() {
		return "workflow-info";
	}

	@Override
	public String getArgumentList(){
		return "[<workflow-file>]";
	}

	@Override
	public String getSynopsis(){
		return "Lists info about workflows. " +
				"The workflow address is read from <workflow-file>, " +
				"if not given, all workflows are shown.";
	}

	@Override
	public String getDescription(){
		return "lists info on workflows.";
	}
	@Override
	public String getCommandGroup(){
		return "Workflow";
	}


	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();
		getOptions().addOption(OptionBuilder.withLongOpt("nofiles")
				.withDescription("Do not list global files (in detailed mode)")
				.isRequired(false)
				.create("N")
				);
		getOptions().addOption(OptionBuilder.withLongOpt("nojobs")
				.withDescription("Do not list jobs (in detailed mode)")
				.isRequired(false)
				.create("j")
				);
	}

	@Override
	public void process(){
		super.process();

		listFiles=detailed & !getBooleanOption("nofiles", "N");

		listJobs=detailed & !getBooleanOption("nojobs", "J");

		if(detailed){
			verbose("Listing jobs: "+listJobs);
			verbose("Listing names of files: "+listFiles);
		}
		try{
			run();
		}catch(Exception e){
			error("",e);
			endProcessing(1);
		}
	}

	//read the workflow address
	protected EndpointReferenceType initEPR()throws Exception{
		String arg=null;
		EndpointReferenceType workflowEPR;

		if(doFilter || getCommandLine().getArgs().length<2)return null;

		arg=getCommandLine().getArgs()[1];

		File wf=new File(arg);
		if(wf.exists())
		{
			workflowEPR=EndpointReferenceType.Factory.parse(wf);
			verbose("Read Workflow EPR from <"+arg+">");
		}
		else{
			//arg is an Address
			workflowEPR=EndpointReferenceType.Factory.newInstance();
			workflowEPR.addNewAddress().setStringValue(arg);
		}
		String url=workflowEPR.getAddress().getStringValue();
		verbose("Checking workflow at "+url);
		return workflowEPR;
	}

	protected void run() throws Exception{
		List<EndpointReferenceType>workflowServices;
		EndpointReferenceType singleEPR=initEPR();
		if(singleEPR!=null){
			WorkflowManagementClient wf=new WorkflowManagementClient(singleEPR.getAddress().getStringValue(),singleEPR,securityProperties);
			if(filterMatch(wf)){
				listOne(wf);
			}
			return;
		}
		else{
			//list all the available workflows
			workflowServices=registry.listAccessibleServices(IWorkflowFactory.PORTTYPE);
		}

		for(EndpointReferenceType epr: workflowServices){
			try{
				verbose("Listing workflows from "+epr.getAddress().getStringValue());
				//TODO workflowsubmission client should do this properties stuff...
				BaseUASClient bc=new BaseUASClient(epr.getAddress().getStringValue(),epr,securityProperties);
				WorkflowFactoryPropertiesDocument wfpd=WorkflowFactoryPropertiesDocument.Factory.parse
						(bc.GetResourcePropertyDocument().getGetResourcePropertyDocumentResponse().newInputStream());
				EndpointReferenceType[]eprs=wfpd.getWorkflowFactoryProperties().getAccessibleWorkflowReferenceArray();
				for(EndpointReferenceType workflowEPR: eprs){
					try{
						WorkflowManagementClient wf=new WorkflowManagementClient(workflowEPR.getAddress().getStringValue(),workflowEPR,securityProperties);
						if(filterMatch(wf)){
							listOne(wf);
						}
					}catch(Exception e){
						verbose("Error accessing workflow at: "+workflowEPR.getAddress().getStringValue());
					}
				}
			}catch(Exception e){
				error("Error accessing workflow service at: "+epr.getAddress().getStringValue(),e);
			}
		}
	}

	protected void listOne(WorkflowManagementClient workflow) throws Exception{
		message(workflow.getEPR().getAddress().getStringValue()+getDetails(workflow));;

		printWSRFProperties(workflow);
	}

	@Override
	protected String getDetails(WorkflowManagementClient workflow)throws Exception{
		if(!detailed)return "";
		StringBuilder details=new StringBuilder();
		String sep=System.getProperty("line.separator");
		details.append(sep);
		details.append("  Status: ").append(workflow.getStatus());

		EndpointReferenceType storageEPR=workflow.getStorageAddress();
		if(storageEPR!=null){
			details.append(sep).append("  Storage: ").append(storageEPR.getAddress().getStringValue());
		}
		EndpointReferenceType tracerEPR=workflow.getTracerAddress();
		if(tracerEPR!=null){
			details.append(sep).append("  Tracer: ").append(tracerEPR.getAddress().getStringValue());
		}
		if(listFiles){
			listFiles(workflow, details);
		}
		try{
			EndpointReferenceType jobList=workflow.getResourcePropertiesDocument().getWorkflowResourceProperties().getJobReferenceEnumeration();
			QName qname=JobReferenceDocument.type.getDocumentElementName();
			EnumerationClient<JobReferenceDocument>enClient=new EnumerationClient<JobReferenceDocument>(jobList,securityProperties,qname);
			details.append(sep).append("  Number of jobs: ").append(enClient.getNumberOfResults());
			if(listJobs){
				listJobs(enClient,details);
			}
		}catch(Exception e){
			logger.error("Can't access job list.",e);
		}
		return details.toString();
	}


	protected void listFiles(WorkflowManagementClient workflow, StringBuilder details){
		try{
			String sep=System.getProperty("line.separator");
			OutputFilesDocument od = OutputFilesDocument.Factory.parse(workflow.getResourceProperty(WorkflowManagement.RPOutputFiles));
			String[] outFiles=od.getOutputFiles().getOutputFileArray();
			for(String file: outFiles){
				details.append(sep).append("  File: ").append(file);
			}
		}catch(Exception e){
			logger.error("Can't retrieve output files",e);
		}
	}

	protected void listJobs(EnumerationClient<JobReferenceDocument> enClient, StringBuilder details){
		String sep=System.getProperty("line.separator");
		try{
			Iterator<JobReferenceDocument>iter=enClient.iterator();
			while(iter.hasNext()){
				details.append(sep).append("  Job: ").append(iter.next().getJobReference().getAddress().getStringValue());
			}
		}catch(Exception e){
			logger.error("Can't process job list.",e);
		}
	}


}
