package de.fzj.unicore.ucc.workflow;

import java.io.IOException;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.cli.OptionBuilder;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.StorageManagement;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.StorageFactoryClient;
import de.fzj.unicore.ucc.actions.ActionBase;
import de.fzj.unicore.ucc.helpers.Location;
import de.fzj.unicore.ucc.helpers.StorageFactoryLister;
import de.fzj.unicore.ucc.util.Builder;
import de.fzj.unicore.ucc.util.FileUploader;

/**
 * contains some common functionality for dealing with the workflow system.
 * <ul>
 *  <li>dealing with storage of workflow data</li>
 *  <li>putting the workflow id into staging URLs</li>
 * </ul>
 * 
 * @author schuller
 */
public class WorkflowSystemSubmission extends ActionBase {

	public static final String OPT_NOFIXFILENAME_LONG="noFilenameFix";
	public static final String OPT_NOFIXFILENAME="F";

	public static final String OPT_STORAGEURL_LONG="storageURL";
	public static final String OPT_STORAGEURL="S";

	protected String siteName;

	protected EndpointReferenceType storageEPR;

	private String storageFactoryURL;

	private String storageURL;

	protected StorageClient sms;

	protected Builder builder;
	
	protected boolean fixFilenames=true;

	protected int lifetime;

	@Override
	@SuppressWarnings("all")
	protected void createOptions() {
		super.createOptions();

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_SITENAME_LONG)
				.withDescription("Site name for submission")
				.withArgName("Vsite")
				.hasArg()
				.isRequired(false)
				.create(OPT_SITENAME)
		);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_NOFIXFILENAME_LONG)
				.withDescription("Do not add the Workflow id to filenames")
				.isRequired(false)
				.create(OPT_NOFIXFILENAME)
		);
		
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_LIFETIME_LONG)
				.withDescription("Initial workflow lifetime in hours")
				.isRequired(false)
				.hasArg()
				.withArgName("InitialLifetime")
				.create(OPT_LIFETIME)
		);

		getOptions().addOption(OptionBuilder.withLongOpt(OPT_FACTORY_LONG)
				.withDescription("URL or site name of storage factory to use")
				.isRequired(false)
				.hasArg()
				.withArgName("StorageFactory")
				.create(OPT_FACTORY)
		);
		
		getOptions().addOption(OptionBuilder.withLongOpt(OPT_STORAGEURL_LONG)
				.withDescription("URL of storage to use for workflow data")
				.withArgName("Storage")
				.hasArg()
				.isRequired(false)
				.create(OPT_STORAGEURL)
		);
	}

	
	@Override
	public void process(){
		super.process();
		siteName=getCommandLine().getOptionValue(OPT_SITENAME);

		storageFactoryURL=getOption(OPT_FACTORY_LONG, OPT_FACTORY);

		storageURL=getOption(OPT_STORAGEURL_LONG, OPT_STORAGEURL);

		if(storageURL!=null && storageFactoryURL!=null){
			error("Error: please, specify either a storage URL or a storage factory URL, but not both!",null);
			endProcessing(1);
		}
		if(getCommandLine().hasOption(OPT_NOFIXFILENAME)){
			fixFilenames=false;
			verbose("Not adding workflow ID to global filenames.");
		}

		lifetime=getNumericOption(OPT_LIFETIME_LONG, OPT_LIFETIME, 24);

	}

	
	/**
	 * create client for a shared space for uploading workflow data, taking
	 * the user's choice into account.
	 *  
	 * <ul>
	 *  <li>if an SMS is given on the command line it is used directly
	 *  <li>the available storage factories are checked
	 *  <li>the available shared storages are checked
	 * <ul>
	 * 
	 * @throws Exception
	 */
	protected void createWorkflowDataStorage()throws Exception{
		if(storageURL!=null){
			useSpecifiedSMS();
		}
		else{
			createStorageOnFactory();
			//if no storage available yet, try shared storage
			if(storageEPR==null){
				useSharedStorage();
			}
		}
		if(storageEPR!=null){
			verbose("Storing workflow data on storage at <"+storageEPR.getAddress().getStringValue()+">");
		}
		else verbose("Could not find a location for storing workflow data.");
	}

	/**
	 * check the StorageFactory instances in the registry, and select one if possible
	 * 
	 * @throws Exception
	 */
	private void createStorageOnFactory()throws Exception{
		for(StorageFactoryClient sc: new StorageFactoryLister(registry,securityProperties,null)){
			String url=sc.getEPR().getAddress().getStringValue();
			if(storageFactoryURL==null){
				if (siteName==null || (siteName!=null && url.contains(siteName))){
					verbose("Using storage factory at <"+url+">");
					createSMS(sc);
					break;
				}
			}
			else if(url.contains(storageFactoryURL)){
				createSMS(sc);
				break;
			}
		}
		
		if(storageFactoryURL==null){
			verbose("No storage factory available.");
		}
	}

	private void createSMS(StorageFactoryClient sfc){
		try{
			storageFactoryURL=sfc.getEPR().getAddress().getStringValue();
			verbose("Storage factory at <"+storageFactoryURL+"> ");
			sms=sfc.createSMS();
			storageEPR=sms.getEPR();
		}catch(Exception ex){
			error("Could not create storage", ex);
			endProcessing(1);
		}
	}

	private void useSharedStorage()throws Exception{
		List<EndpointReferenceType> storages=registry.listAccessibleServices(StorageManagement.SMS_PORT);
		if(storages.size()==0){
			verbose("No shared storages found in registry.");
			return;
		}
		else {
			storageEPR=storages.get(0);
			//check site name
			if(siteName!=null && storages.size()>1){
				for(EndpointReferenceType e: storages){
					if(e.getAddress().getStringValue().contains(siteName)){
						verbose("Using storage at site "+siteName);
						storageEPR=e;
						break;
					}
				}
			}
		}
		storageURL=storageEPR.getAddress().getStringValue();
		sms=new StorageClient(storageEPR,securityProperties);
	}

	private void useSpecifiedSMS()throws Exception{
		verbose("Storage at <"+storageURL+"> ");
		storageURL=resolveU6Location(storageURL);
		storageEPR=EndpointReferenceType.Factory.newInstance();
		storageEPR.addNewAddress().setStringValue(storageURL);
		sms=new StorageClient(storageEPR,securityProperties);
	}

	/**
	 * upload local files to the shared storage
	 * 
	 * @param builder - a {@link Builder} containing the local imports
	 * @param uniqueID - a unique identifier to use as a storage directory
	 * @throws IOException in case of problems to upload the data
	 */
	protected void uploadLocalData(Builder builder, String uniqueID)throws IOException{
		GlobalFileImporter gif=new GlobalFileImporter(registry,securityProperties,uniqueID,this);
		gif.setStorage(sms);
		for(FileUploader imp: builder.getImports()){
			String to=fixFilenames?writeNewIDs(imp.getTo(), uniqueID):imp.getTo();
			gif.importFile(imp.getFrom(), to);
		}
	}

	/**
	 * pre-pends the unique workflow ID to logical filenames, e.g.
	 * $WORKFLOW_ID -> UUID
	 */
	public static String writeNewIDs(String s, String newID){
		return s.replaceAll("\\$WORKFLOW_ID|\\$\\{WORKFLOW_ID\\}", newID);
	}

	/**
	 * resolves u6:// urls into their real value
	 */
	static Pattern p=Pattern.compile("(u6://[^<|\\$]*)|(unicore6://[^<|\\$]*)");

	public String resolveU6URLs(String s){
		String result=s;
		while(true){
			Matcher m=p.matcher(result);
			if(m.find()){
				String seq=m.group();
				Location l=new Location(seq,registry,securityProperties,this);
				String resolved=l.getUnicore6URI();
				verbose("Resolved <"+seq+"> as "+resolved);
				result=m.replaceFirst(resolved);
			}
			else break;
		}
		return result;
	}

	public String resolveU6Location(String s){
		while(true){
			Matcher m=p.matcher(s);
			if(m.find()){
				String seq=m.group();
				Location l=new Location(seq,registry,securityProperties,this);
				String resolved=l.getSmsEpr();
				verbose("Resolved <"+seq+"> as "+resolved);
				return resolved;
			}
			else break;
		}
		return s;
	}

}
